"""Runway variables."""
import logging
import re
from typing import (TYPE_CHECKING,  # noqa: F401 pylint: disable=W
                    Any, Dict, Iterable, Iterator, List, Optional, Set, Type,
                    Union, cast)

from six import string_types

from .cfngin.exceptions import (FailedLookup, FailedVariableLookup,
                                InvalidLookupCombination,
                                InvalidLookupConcatenation, UnknownLookupType,
                                UnresolvedVariable, UnresolvedVariableValue)
from .cfngin.lookups.registry import CFNGIN_LOOKUP_HANDLERS
from .lookups.handlers.base import \
    LookupHandler  # noqa: F401 pylint: disable=unused-import
from .lookups.registry import RUNWAY_LOOKUP_HANDLERS

# python2 supported pylint sees this is cyclic even though its only for type checking
# pylint: disable=cyclic-import
if TYPE_CHECKING:
    from .config import VariablesDefinition  # noqa: F401 pylint: disable=unused-import


LOGGER = logging.getLogger('runway')


def resolve_variables(variables, context, provider):
    """Given a list of variables, resolve all of them.

    Args:
        variables (List[:class:`Variable`]): List of variables.
        context (:class:`runway.cfngin.context.Context`): CFNgin context.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`): Subclass
            of the base provider.

    """
    for variable in variables:
        variable.resolve(context, provider)


class Variable(object):
    """Represents a variable provided to a Runway directive."""

    def __init__(self, name, value, variable_type='cfngin'):
        # type: (str, Any, str) -> None
        """Initialize class.

        Args:
            name: Name of the variable (directive/key).
            value: The variable itself.

        """
        LOGGER.debug('Initalized variable "%s".', name)
        self.name = name
        self._raw_value = value
        self._value = VariableValue.parse(value, variable_type)

    @property
    def dependencies(self):
        # () -> Set[str]
        """Stack names that this variable depends on.

        Returns:
            Set[str]: Stack names that this variable depends on.

        """
        return self._value.dependencies

    @property
    def resolved(self):
        # type: () -> bool
        """Boolean for whether the Variable has been resolved.

        Variables only need to be resolved if they contain lookups.

        """
        return self._value.resolved

    @property
    def value(self):
        # type: () -> Any
        """Return the current value of the Variable."""
        try:
            return self._value.value
        except UnresolvedVariableValue:
            raise UnresolvedVariable("<unknown>", self)
        except InvalidLookupConcatenation as err:
            raise InvalidLookupCombination(err.lookup, err.lookups, self)

    def resolve(self, context, provider=None, variables=None, **kwargs):
        # type: (Any, Any, 'Optional[VariablesDefinition]', Any) -> None
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.

        """
        try:
            self._value.resolve(context, provider=provider,
                                variables=variables, **kwargs)
        except FailedLookup as err:
            raise FailedVariableLookup(self.name, err.lookup, err.error)

    def get(self, key, default=None):
        # type: (Any, Any) -> Any
        """Implement evaluation of self.get.

        Args:
            key: Attribute name to return the value for.
            default: Value to return if attribute is not found.

        """
        return getattr(self.value, key, default)

    def __repr__(self):
        # type: () -> str
        """Return object representation."""
        return 'Variable<{}={}>'.format(self.name, self._raw_value)


class VariableValue(object):
    """Syntax tree base class to parse variable values."""

    @property
    def dependencies(self):
        # () -> Set[]
        """Stack names that this variable depends on."""
        return set()

    @property
    def resolved(self):
        # type: () -> bool
        """Use to check if the variable value has been resolved.

        Should be implimented in subclasses.

        """
        raise NotImplementedError

    @property
    def simplified(self):
        # type: () -> Any
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        Should be implimented in subclasses where applicable.

        """
        return self

    @property
    def value(self):
        # type: () -> Any
        """Value of the variable. Can be resolved or unresolved.

        Should be implimented in subclasses.

        """
        raise NotImplementedError

    def resolve(self, context, provider=None, variables=None, **kwargs):
        # type: (Any, Any, 'Optional[VariablesDefinition]', Any) -> None
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.

        """

    @classmethod
    def parse(cls, input_object, variable_type='cfngin'):
        # type: (Any, str) -> Any
        """Parse complex variable structures using type appropriate subclasses.

        Args:
            input_object: The objected defined as the value of a variable.

        """
        if isinstance(input_object, list):
            return VariableValueList.parse(input_object, variable_type)
        if isinstance(input_object, dict):
            return VariableValueDict.parse(input_object, variable_type)
        if not isinstance(input_object, string_types):
            return VariableValueLiteral(input_object)

        tokens = VariableValueConcatenation([
            VariableValueLiteral(t)
            for t in re.split(r'(\$\{|\}|\s+)', input_object)  # ${ or space or }
        ])

        opener = '${'
        closer = '}'

        while True:
            last_open = None
            next_close = None
            for i, tok in enumerate(tokens):
                if not isinstance(tok, VariableValueLiteral):
                    continue

                if tok.value == opener:
                    last_open = i
                    next_close = None
                if (
                        last_open is not None and
                        tok.value == closer and
                        next_close is None
                ):
                    next_close = i

            if next_close is not None:
                lookup_data = VariableValueConcatenation(
                    tokens[(cast(int, last_open) + len(opener) + 1):next_close]
                )
                lookup = VariableValueLookup(
                    lookup_name=tokens[cast(int, last_open) + 1],
                    lookup_data=lookup_data,
                    variable_type=variable_type
                )
                tokens[last_open:(next_close + 1)] = [lookup]
            else:
                break

        return tokens.simplified

    def __iter__(self):
        # type: () -> Iterable
        """How the object is iterated.

        Should be implimented in subclasses.

        """
        raise NotImplementedError

    def __repr__(self):
        # type: () -> str
        """Return object representation.

        Should be implimented in subclasses.

        """
        raise NotImplementedError


class VariableValueLiteral(VariableValue):
    """The literal value of a variable as provided."""

    def __init__(self, value):
        # type: (Any) -> None
        """Initialize class."""
        self._value = value

    @property
    def resolved(self):
        # type: () -> bool
        """Use to check if the variable value has been resolved.

        The ValueLiteral will always appear as resolved because it does
        not "resolve" since it is the literal definition of the value.

        """
        return True

    @property
    def value(self):
        # type: () -> Any
        """Value of the variable."""
        return self._value

    def __iter__(self):
        # type: () -> Iterable[Any]
        """How the object is iterated."""
        yield self

    def __repr__(self):
        # type: () -> str
        """Return object representation."""
        return 'Literal<{}>'.format(repr(self._value))


class VariableValueList(VariableValue, list):
    """A list variable value."""

    @property
    def dependencies(self):
        # () -> Set[str]
        """Stack names that this variable depends on."""
        deps = set()
        for item in self:
            deps.update(item.dependencies)
        return deps

    @property
    def resolved(self):
        # type: () -> bool
        """Use to check if the variable value has been resolved."""
        accumulator = True
        for item in self:
            accumulator = accumulator and item.resolved
        return accumulator

    @property
    def simplified(self):
        # type: () -> List[VariableValue]
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        """
        return [
            item.simplified
            for item in self
        ]

    @property
    def value(self):
        # type: () -> List[Any]
        """Value of the variable. Can be resolved or unresolved."""
        return [
            item.value
            for item in self
        ]

    def resolve(self, context, provider=None, variables=None, **kwargs):
        # type: (Any, Any, 'Optional[VariablesDefinition]', Any) -> None
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.

        """
        for item in self:
            item.resolve(context, variables=variables, **kwargs)

    @classmethod
    def parse(cls, input_object, variable_type='cfngin'):
        # type: (Any, str) -> VariableValueList
        """Parse list variable structure.

        Args:
            input_object: The objected defined as the value of a variable.

        """
        acc = [
            VariableValue.parse(obj, variable_type)
            for obj in input_object
        ]
        return cls(acc)

    def __iter__(self):
        # type: () -> Iterator[Any]
        """How the object is iterated."""
        return list.__iter__(self)

    def __repr__(self):
        # type: () -> str
        """Return object representation."""
        return 'List[{}]'.format(', '.join([repr(value)
                                            for value in self]))


class VariableValueDict(VariableValue, dict):
    """A dict variable value."""

    @property
    def dependencies(self):
        # () -> Set[str]
        """Stack names that this variable depends on."""
        deps = set()
        for item in self:
            deps.update(item.dependencies)
        return deps

    @property
    def resolved(self):
        # type: () -> bool
        """Use to check if the variable value has been resolved."""
        accumulator = True
        for item in self.values():
            accumulator = accumulator and item.resolved
        return accumulator

    @property
    def simplified(self):
        # type: () -> Dict[str, VariableValue]
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        """
        return {
            k: v.simplified
            for k, v in self.items()
        }

    @property
    def value(self):
        # type: () -> Dict[str, Any]
        """Value of the variable. Can be resolved or unresolved."""
        return {
            k: v.value
            for k, v in self.items()
        }

    def resolve(self, context, provider=None, variables=None, **kwargs):
        # type: (Any, Any, 'Optional[VariablesDefinition]', Any) -> None
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.

        """
        for item in self.values():
            item.resolve(context, variables=variables, **kwargs)

    @classmethod
    def parse(cls, input_object, variable_type='cfngin'):
        # type: (Any, str) -> VariableValueDict
        """Parse list variable structure.

        Args:
            input_object: The objected defined as the value of a variable.

        """
        acc = {
            k: VariableValue.parse(v, variable_type)
            for k, v in input_object.items()
        }
        return cls(acc)

    def __iter__(self):
        # type: () -> Iterator[Any]
        """How the object is iterated."""
        return dict.__iter__(self)

    def __repr__(self):
        # type: () -> str
        """Return object representation."""
        return 'Dict[{}]'.format(', '.join([
            "{}={}".format(k, repr(v)) for k, v in self.items()
        ]))


class VariableValueConcatenation(VariableValue, list):
    """A concatinated variable value."""

    @property
    def dependencies(self):
        """Stack names that this variable depends on."""
        deps = set()
        for item in self:
            deps.update(item.dependencies)
        return deps

    @property
    def resolved(self):
        # type: () -> bool
        """Use to check if the variable value has been resolved."""
        accumulator = True
        for item in self:
            accumulator = bool(accumulator and item.resolved)
        return accumulator

    @property
    def simplified(self):
        # type: () -> Union[Type[VariableValue], VariableValueConcatenation, VariableValueLiteral]
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        """
        concat = []  # type: List[Type[VariableValue]]
        for item in self:
            if (
                    isinstance(item, VariableValueLiteral) and
                    item.value == ''
            ):
                pass

            elif (
                    isinstance(item, VariableValueLiteral) and concat and
                    isinstance(concat[-1], VariableValueLiteral)
            ):
                # join the literals together
                concat[-1] = VariableValueLiteral(
                    concat[-1].value + item.value
                )

            elif isinstance(item, VariableValueConcatenation):
                # flatten concatenations
                concat.extend(item.simplified)

            else:
                concat.append(cast(Type[VariableValue], item.simplified))

        if not concat:
            return VariableValueLiteral('')
        if len(concat) == 1:
            return concat[0]
        return VariableValueConcatenation(concat)

    @property
    def value(self):
        # type: () -> Any
        """Value of the variable. Can be resolved or unresolved."""
        if len(self) == 1:
            return self[0].value

        values = []  # type: List[str]
        for value in self:
            resolved_value = value.value
            if not isinstance(resolved_value, string_types):
                raise InvalidLookupConcatenation(value, self)
            values.append(resolved_value)
        return ''.join(values)

    def resolve(self, context, provider=None, variables=None, **kwargs):
        # type: (Any, Any, 'Optional[VariablesDefinition]', Any) -> None
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.

        """
        for value in self:
            value.resolve(context, variables=variables, **kwargs)

    def __iter__(self):
        # type: () -> Iterator[Type[VariableValue]]
        """How the object is iterated."""
        return list.__iter__(self)

    def __repr__(self):
        # type: () -> str
        """Return object representation."""
        return 'Concatenation[{}]'.format(
            ', '.join([repr(value) for value in self])
        )


class VariableValueLookup(VariableValue):
    """A lookup variable value."""

    # flake8/pylint fight over the indent here on python2
    # TODO remove "disable=bad-continuation" when dropping python2
    def __init__(self,
                 lookup_name,  # type: VariableValueLiteral
                 lookup_data,  # type: VariableValue
                 handler=None,  # type: Optional[Type[LookupHandler]]
                 variable_type='cfngin'  # type: str
                 ):  # pylint: disable=bad-continuation
        # type: (...) -> None
        """Initialize class.

        Args:
            lookup_name: Name of the invoked lookup
            lookup_data: Data portion of the lookup

        """
        self._resolved = False
        self._value = None

        self.lookup_name = lookup_name

        if isinstance(lookup_data, string_types):
            lookup_data = VariableValueLiteral(lookup_data)
        self.lookup_data = lookup_data

        if handler is None:
            lookup_name_resolved = lookup_name.value
            try:
                if variable_type == 'cfngin':
                    handler = cast(
                        Type[LookupHandler],
                        CFNGIN_LOOKUP_HANDLERS[lookup_name_resolved]
                    )
                elif variable_type == 'runway':
                    handler = cast(
                        Type[LookupHandler],
                        RUNWAY_LOOKUP_HANDLERS[lookup_name_resolved]
                    )
                else:
                    raise ValueError(
                        'Variable type must be one of "cfngin" or "runway"'
                    )
            except KeyError:
                raise UnknownLookupType(lookup_name_resolved)
        self.handler = handler

    @property
    def dependencies(self):
        # () -> Set[str]
        """Stack names that this variable depends on."""
        if isinstance(self.handler, type):
            return self.handler.dependencies(self.lookup_data)
        return set()

    @property
    def resolved(self):
        # type: () -> bool
        """Use to check if the variable value has been resolved."""
        return self._resolved

    @property
    def simplified(self):
        # type: () -> VariableValueLookup
        """Return a simplified version of the value.

        This can be used to concatenate two literals into one literal or
        flatten nested concatenations.

        """
        return self

    @property
    def value(self):
        # type: () -> Any
        """Value of the variable. Can be resolved or unresolved."""
        if self._resolved:
            return self._value
        raise UnresolvedVariableValue(self)

    def resolve(self, context, provider=None, variables=None, **kwargs):
        # type: (Any, Any, 'Optional[VariablesDefinition]', Any) -> None
        """Resolve the variable value.

        Args:
            context: The current context object.
            provider: Subclass of the base provider.
            variables: Object containing variables passed to Runway.

        Raises:
            FailedLookup: A lookup failed for any reason.

        """
        self.lookup_data.resolve(context, variables=variables, **kwargs)
        try:
            if isinstance(self.handler, type):
                result = self.handler.handle(value=self.lookup_data.value,
                                             context=context,
                                             provider=provider,
                                             variables=variables,
                                             **kwargs)
            else:
                result = self._resolve_legacy(context, provider)
            return self._resolve(result)
        except Exception as err:
            if isinstance(err, TypeError):
                # handle lookups that don't accept all the args we want
                # to pass to it
                LOGGER.debug('Encountered %s: %s - trying legacy resolver',
                             type(err), err)
                try:
                    return self._resolve(self._resolve_legacy(context,
                                                              provider))
                except Exception as err2:
                    raise FailedLookup(self, err2)
            raise FailedLookup(self, err)

    def _resolve(self, value):
        # type: (Any) -> None
        """Set _value and _resolved from the result of resolve()."""
        self._value = value
        self._resolved = True

    def _resolve_legacy(self, context, provider):
        """Resolve legacy lookups.

        Stacker style custom lookups only take 3 args (value, provider,
        context). In combining CFNgin and Runway Variable/LookupHandler classes
        we need to be able to pass more arguments. The built-in lookups have
        been updated but, any custom lookups designed for Stacker would fail
        when trying to pass the arguments that now get passed. That is where
        this method comes it. It can handle legacy Stacker function lookups
        and those that don't support accept more then 3 args.

        TODO:
            Remove during the next major release.

        """
        if isinstance(self.handler, type):
            import warnings
            warn_msg = ('Old style lookup in use. Please upgrade to use '
                        'the new style of Lookups that accepts '
                        '"**kwargs".')
            LOGGER.warning(warn_msg)
            warnings.warn(warn_msg, DeprecationWarning, stacklevel=2)
            return self.handler.handle(value=self.lookup_data.value,
                                       context=context,
                                       provider=provider)
        return self.handler(value=self.lookup_data.value,
                            context=context,
                            provider=provider)

    def __iter__(self):
        # type: () -> Iterable
        """How the object is iterated."""
        yield self

    def __repr__(self):
        # type: () -> str
        """Return object representation."""
        if self._resolved:
            return 'Lookup<{r} ({t} {d})>'.format(
                r=self._value,
                t=self.lookup_name,
                d=repr(self.lookup_data),
            )
        return 'Lookup<{t} {d}>'.format(
            t=self.lookup_name,
            d=repr(self.lookup_data),
        )

    def __str__(self):
        # type: () -> str
        """Object displayed as a string."""
        return '${{{type} {data}}}'.format(
            type=self.lookup_name.value,
            data=self.lookup_data.value,
        )



"""Tests for lookup handler base class."""
from unittest import TestCase

from runway.lookups.handlers.base import LookupHandler


class TestLookupHandler(TestCase):
    """Tests for LookupHandler."""

    def test_abstract_handle(self):
        """Handle should not be implimented."""
        with self.assertRaises(NotImplementedError):
            LookupHandler.handle(None, None)

    def test_parse(self):
        """Basic value parsing."""
        expected_query = 'my_query'

        result_query, result_args = LookupHandler.parse(expected_query)

        self.assertEqual(result_query, expected_query)
        self.assertEqual(result_args, {})

    def test_parse_args(self):
        """Parse query and args from value."""
        expected_args = {
            'key1': 'val1'
        }
        expected_query = 'my_query'
        value = '{}::{}'.format(expected_query, ','.join([
            '{}={}'.format(key, val) for key, val in expected_args.items()
        ]))

        result_query, result_args = LookupHandler.parse(value)

        self.assertEqual(result_query, expected_query)
        self.assertEqual(result_args, expected_args)

    def test_transform_bool_to_bool(self):
        """Bool should be returned as is."""
        result_true = LookupHandler.transform(True, to_type='bool')
        result_false = LookupHandler.transform(False, to_type='bool')

        self.assertTrue(result_true)
        self.assertFalse(result_false)

    def test_transform_str_to_bool(self):
        """String should be transformed using strtobool."""
        result_true = LookupHandler.transform('true', to_type='bool')
        result_false = LookupHandler.transform('false', to_type='bool')

        self.assertTrue(result_true)
        self.assertFalse(result_false)

    def test_transform_type_check(self):
        """Transform to bool type check."""
        with self.assertRaises(TypeError, msg='dict should raise an error'):
            LookupHandler.transform({'key1': 'val1'}, to_type='bool')

        with self.assertRaises(TypeError, msg='list should raise an error'):
            LookupHandler.transform(['li1'], to_type='bool')

        with self.assertRaises(TypeError, msg='number should raise an error'):
            LookupHandler.transform(10, to_type='bool')

        with self.assertRaises(TypeError, msg='float should raise an error'):
            LookupHandler.transform(10.0, to_type='bool')

        with self.assertRaises(TypeError, msg='NoneType should raise an error'):
            LookupHandler.transform(None, to_type='bool')

    def test_transform_str_direct(self):
        """Test types that are directly transformed to strings."""
        self.assertEqual(LookupHandler.transform('test', 'str'), 'test')
        self.assertEqual(LookupHandler.transform({'key1': 'val1'}, 'str'),
                         "{'key1': 'val1'}")
        self.assertEqual(LookupHandler.transform(True, 'str'), 'True')

    def test_transform_str_list(self):
        """Test list type joined to create string."""
        self.assertEqual(
            LookupHandler.transform(['val1', 'val2'], to_type='str'),
            'val1,val2'
        )
        self.assertEqual(
            LookupHandler.transform(set(['val', 'val']), to_type='str'),
            'val'
        )
        self.assertEqual(
            LookupHandler.transform(('val1', 'val2'), to_type='str'),
            'val1,val2'
        )

    def test_transform_str_list_delimiter(self):
        """Test list to string with a specified delimiter."""
        self.assertEqual(
            LookupHandler.transform(['val1', 'val2'], to_type='str',
                                    delimiter='|'),
            'val1|val2'
        )



"""Test destroying stacks using tags."""
from integration_tests.test_moduletags.test_moduletags import ModuleTags


class TestDestroy(ModuleTags):
    """Tests destroy using module tags."""

    TEST_NAME = __name__

    def init(self):
        """Initialize test."""
        self.clean()
        self.copy_sampleapp()

    def run(self):
        """Run tests."""
        self.init()
        self.runway_cmd('deploy', [])
        assert self.runway_cmd('destroy', ['app1']) == 0
        self.check_stacks([str(num) for num in range(2, 7)])
        assert self.runway_cmd('destroy', ['app2']) == 0
        self.check_stacks([str(num) for num in range(3, 7)])
        self.runway_cmd('deploy', [])
        assert self.runway_cmd('destroy', ['group-1-4']) == 0
        self.check_stacks(['2', '3', '5', '6'])

    def teardown(self):
        """Teardown any created resources."""
        self.runway_cmd('destroy', [])
        self.clean()



"""CFNgin test."""
# flake8: noqa
# pylint: disable=invalid-name
from os.path import basename

from integration_tests.test_cfngin.test_cfngin import Cfngin

FILE_BASENAME = '.'.join(basename(__file__).split('.')[:-1])


class TestRollbackWithDependent(Cfngin):
    """Test CFNgin stack rollback with a dependent.

    Requires valid AWS credentials.

    This will take a bit of time to run (~5 min) due to needing to create
    a failed stack, allow it to rollback, and redeploy.

    """

    REQUIRED_FIXTURE_FILES = [FILE_BASENAME + '.yaml']
    TEST_NAME = __name__

    def run(self):
        """Run the test."""
        self.copy_fixtures()
        code, _stdout, stderr = self.runway_cmd('deploy')
        assert code != 0, 'exit code should be non-zero'
        expected_lines = [
            'dependent-rollback-parent: submitted (creating new stack)',
            # the suffix of the below log message can very based on when
            # CFN is polled b/c of how fast the test stack is
            'dependent-rollback-parent: failed',
            'dependent-rollback-child: failed (dependency has failed)',
            'The following steps failed: dependent-rollback-parent, dependent-rollback-child'
        ]
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def teardown(self):
        """Teardown any created resources and delete files."""
        self.runway_cmd('destroy')
        self.cleanup_fixtures()



#!/usr/bin/env python
"""Module with app build pipeline."""
from __future__ import print_function

from os import path

from troposphere import (
    AWSHelperFn, AccountId, Join, Output, Partition, Region, awslambda,
    codecommit, codebuild, codepipeline, events, iam, s3
)

import awacs.awslambda
import awacs.codebuild
import awacs.codecommit
import awacs.ecr
import awacs.logs
import awacs.s3
import awacs.ssm
from awacs.aws import Allow, PolicyDocument, Statement
from awacs.helpers.trust import make_simple_assume_policy

from stacker.lookups.handlers.file import parameterized_codec
from stacker.blueprints.base import Blueprint
from stacker.blueprints.variables.types import CFNString

AWS_LAMBDA_DIR = path.join(path.dirname(path.realpath(__file__)),
                           'aws_lambda')
IAM_ARN_PREFIX = 'arn:aws:iam::aws:policy/service-role/'


class Pipeline(Blueprint):
    """Stacker blueprint for app building components."""

    cleanup_ecr_src = parameterized_codec(
        open(path.join(AWS_LAMBDA_DIR, 'cleanup_ecr.py'), 'r').read(),
        False  # disable base64 encoding
    )

    build_proj_spec = parameterized_codec(
        open(path.join(path.dirname(path.realpath(__file__)),
                       'build_project_buildspec.yml'), 'r').read(),
        False  # disable base64 encoding
    )

    VARIABLES = {
        'ECRCleanupLambdaFunction': {'type': AWSHelperFn,
                                     'description': 'Lambda function code',
                                     'default': cleanup_ecr_src},
        'BuildProjectBuildSpec': {'type': AWSHelperFn,
                                  'description': 'Inline buildspec code',
                                  'default': build_proj_spec},
        'AppPrefix': {'type': CFNString,
                      'description': 'Application prefix (for roles, etc)'},
        'EcrRepoName': {'type': CFNString,
                        'description': 'Name of ECR repo'},
        'RolePermissionsBoundaryName': {'type': CFNString,
                                        'description': 'Roles\' boundary '
                                                       'name'},
    }

    def create_template(self):
        """Create template (main function called by Stacker)."""
        template = self.template
        variables = self.get_variables()
        template.set_version('2010-09-09')
        template.set_description('App - Build Pipeline')

        # Resources
        boundary_arn = Join('',
                            ['arn:',
                             Partition,
                             ':iam::',
                             AccountId,
                             ':policy/',
                             variables['RolePermissionsBoundaryName'].ref])

        # Repo image limit is 1000 by default; this lambda function will prune
        # old images
        image_param_path = Join('',
                                ['/',
                                 variables['AppPrefix'].ref,
                                 '/current-hash'])
        image_param_arn = Join('',
                               ['arn:',
                                Partition,
                                ':ssm:',
                                Region,
                                ':',
                                AccountId,
                                ':parameter',
                                image_param_path])
        ecr_repo_arn = Join('',
                            ['arn:',
                             Partition,
                             ':ecr:',
                             Region,
                             ':',
                             AccountId,
                             ':repository/',
                             variables['EcrRepoName'].ref])
        cleanuplambdarole = template.add_resource(
            iam.Role(
                'CleanupLambdaRole',
                AssumeRolePolicyDocument=make_simple_assume_policy(
                    'lambda.amazonaws.com'
                ),
                ManagedPolicyArns=[
                    IAM_ARN_PREFIX + 'AWSLambdaBasicExecutionRole'
                ],
                PermissionsBoundary=boundary_arn,
                Policies=[
                    iam.Policy(
                        PolicyName=Join('', [variables['AppPrefix'].ref,
                                             '-ecrcleanup']),
                        PolicyDocument=PolicyDocument(
                            Version='2012-10-17',
                            Statement=[
                                Statement(
                                    Action=[awacs.ssm.GetParameter],
                                    Effect=Allow,
                                    Resource=[
                                        image_param_arn
                                    ]
                                ),
                                Statement(
                                    Action=[
                                        awacs.ecr.DescribeImages,
                                        awacs.ecr.BatchDeleteImage
                                    ],
                                    Effect=Allow,
                                    Resource=[ecr_repo_arn]
                                )
                            ]
                        )
                    )
                ]
            )
        )
        cleanupfunction = template.add_resource(
            awslambda.Function(
                'CleanupFunction',
                Description='Cleanup stale ECR images',
                Code=awslambda.Code(
                    ZipFile=variables['ECRCleanupLambdaFunction']
                ),
                Environment=awslambda.Environment(
                    Variables={
                        'ECR_REPO_NAME': variables['EcrRepoName'].ref,
                        'SSM_PARAM': image_param_path
                    }
                ),
                Handler='index.handler',
                Role=cleanuplambdarole.get_att('Arn'),
                Runtime='python3.6',
                Timeout=120
            )
        )
        cleanuprule = template.add_resource(
            events.Rule(
                'CleanupRule',
                Description='Regularly invoke CleanupFunction',
                ScheduleExpression='rate(7 days)',
                State='ENABLED',
                Targets=[
                    events.Target(
                        Arn=cleanupfunction.get_att('Arn'),
                        Id='CleanupFunction'
                    )
                ]
            )
        )
        template.add_resource(
            awslambda.Permission(
                'AllowCWLambdaInvocation',
                FunctionName=cleanupfunction.ref(),
                Action=awacs.awslambda.InvokeFunction.JSONrepr(),
                Principal='events.amazonaws.com',
                SourceArn=cleanuprule.get_att('Arn')
            )
        )

        appsource = template.add_resource(
            codecommit.Repository(
                'AppSource',
                RepositoryName=Join('-',
                                    [variables['AppPrefix'].ref,
                                     'source'])
            )
        )
        for i in ['Name', 'Arn']:
            template.add_output(Output(
                "AppRepo%s" % i,
                Description="%s of app source repo" % i,
                Value=appsource.get_att(i)
            ))

        bucket = template.add_resource(
            s3.Bucket(
                'Bucket',
                AccessControl=s3.Private,
                LifecycleConfiguration=s3.LifecycleConfiguration(
                    Rules=[
                        s3.LifecycleRule(
                            NoncurrentVersionExpirationInDays=90,
                            Status='Enabled'
                        )
                    ]
                ),
                VersioningConfiguration=s3.VersioningConfiguration(
                    Status='Enabled'
                )
            )
        )
        template.add_output(Output(
            'PipelineBucketName',
            Description='Name of pipeline bucket',
            Value=bucket.ref()
        ))

        # This list must be kept in sync between the CodeBuild project and its
        # role
        build_name = Join('', [variables['AppPrefix'].ref, '-build'])

        build_role = template.add_resource(
            iam.Role(
                'BuildRole',
                AssumeRolePolicyDocument=make_simple_assume_policy(
                    'codebuild.amazonaws.com'
                ),
                PermissionsBoundary=boundary_arn,
                Policies=[
                    iam.Policy(
                        PolicyName=Join('', [build_name, '-policy']),
                        PolicyDocument=PolicyDocument(
                            Version='2012-10-17',
                            Statement=[
                                Statement(
                                    Action=[awacs.s3.GetObject],
                                    Effect=Allow,
                                    Resource=[
                                        Join('',
                                             [bucket.get_att('Arn'),
                                              '/*'])
                                    ]
                                ),
                                Statement(
                                    Action=[awacs.ecr.GetAuthorizationToken],
                                    Effect=Allow,
                                    Resource=['*']
                                ),
                                Statement(
                                    Action=[
                                        awacs.ecr.BatchCheckLayerAvailability,
                                        awacs.ecr.BatchGetImage,
                                        awacs.ecr.CompleteLayerUpload,
                                        awacs.ecr.DescribeImages,
                                        awacs.ecr.GetDownloadUrlForLayer,
                                        awacs.ecr.InitiateLayerUpload,
                                        awacs.ecr.PutImage,
                                        awacs.ecr.UploadLayerPart
                                    ],
                                    Effect=Allow,
                                    Resource=[ecr_repo_arn]
                                ),
                                Statement(
                                    Action=[awacs.ssm.GetParameter,
                                            awacs.ssm.PutParameter],
                                    Effect=Allow,
                                    Resource=[
                                        image_param_arn
                                    ]
                                ),
                                Statement(
                                    Action=[
                                        awacs.logs.CreateLogGroup,
                                        awacs.logs.CreateLogStream,
                                        awacs.logs.PutLogEvents
                                    ],
                                    Effect=Allow,
                                    Resource=[
                                        Join(
                                            '',
                                            [
                                                'arn:',
                                                Partition,
                                                ':logs:',
                                                Region,
                                                ':',
                                                AccountId,
                                                ':log-group:/aws/codebuild/',
                                                build_name
                                            ] + x
                                        ) for x in [[':*'], [':*/*']]
                                    ]
                                )
                            ]
                        )
                    )
                ]
            )
        )

        buildproject = template.add_resource(
            codebuild.Project(
                'BuildProject',
                Artifacts=codebuild.Artifacts(
                    Type='CODEPIPELINE'
                ),
                Environment=codebuild.Environment(
                    ComputeType='BUILD_GENERAL1_SMALL',
                    EnvironmentVariables=[
                        codebuild.EnvironmentVariable(
                            Name='AWS_DEFAULT_REGION',
                            Type='PLAINTEXT',
                            Value=Region
                        ),
                        codebuild.EnvironmentVariable(
                            Name='AWS_ACCOUNT_ID',
                            Type='PLAINTEXT',
                            Value=AccountId
                        ),
                        codebuild.EnvironmentVariable(
                            Name='IMAGE_REPO_NAME',
                            Type='PLAINTEXT',
                            Value=variables['EcrRepoName'].ref
                        ),
                    ],
                    Image='aws/codebuild/docker:18.09.0',
                    Type='LINUX_CONTAINER'
                ),
                Name=build_name,
                ServiceRole=build_role.get_att('Arn'),
                Source=codebuild.Source(
                    Type='CODEPIPELINE',
                    BuildSpec=variables['BuildProjectBuildSpec']
                )
            )
        )

        pipelinerole = template.add_resource(
            iam.Role(
                'PipelineRole',
                AssumeRolePolicyDocument=make_simple_assume_policy(
                    'codepipeline.amazonaws.com'
                ),
                PermissionsBoundary=boundary_arn,
                Policies=[
                    iam.Policy(
                        PolicyName=Join('', [build_name, '-pipeline-policy']),
                        PolicyDocument=PolicyDocument(
                            Version='2012-10-17',
                            Statement=[
                                Statement(
                                    Action=[awacs.codecommit.GetBranch,
                                            awacs.codecommit.GetCommit,
                                            awacs.codecommit.UploadArchive,
                                            awacs.codecommit.GetUploadArchiveStatus,  # noqa
                                            awacs.codecommit.CancelUploadArchive],  # noqa
                                    Effect=Allow,
                                    Resource=[appsource.get_att('Arn')]
                                ),
                                Statement(
                                    Action=[awacs.s3.GetBucketVersioning],
                                    Effect=Allow,
                                    Resource=[bucket.get_att('Arn')]
                                ),
                                Statement(
                                    Action=[awacs.s3.GetObject,
                                            awacs.s3.PutObject],
                                    Effect=Allow,
                                    Resource=[
                                        Join('',
                                             [bucket.get_att('Arn'),
                                              '/*'])
                                    ]
                                ),
                                Statement(
                                    Action=[
                                        awacs.codebuild.BatchGetBuilds,
                                        awacs.codebuild.StartBuild
                                    ],
                                    Effect=Allow,
                                    Resource=[
                                        buildproject.get_att('Arn')
                                    ]
                                )
                            ]
                        )
                    )
                ]
            )
        )

        template.add_resource(
            codepipeline.Pipeline(
                'Pipeline',
                ArtifactStore=codepipeline.ArtifactStore(
                    Location=bucket.ref(),
                    Type='S3'
                ),
                Name=build_name,
                RoleArn=pipelinerole.get_att('Arn'),
                Stages=[
                    codepipeline.Stages(
                        Name='Source',
                        Actions=[
                            codepipeline.Actions(
                                Name='CodeCommit',
                                ActionTypeId=codepipeline.ActionTypeId(
                                    Category='Source',
                                    Owner='AWS',
                                    Provider='CodeCommit',
                                    Version='1'
                                ),
                                Configuration={
                                    'RepositoryName': appsource.get_att('Name'),  # noqa
                                    'BranchName': 'master'
                                },
                                OutputArtifacts=[
                                    codepipeline.OutputArtifacts(
                                        Name='CodeCommitRepo'
                                    )
                                ]
                            ),
                        ]
                    ),
                    codepipeline.Stages(
                        Name='Build',
                        Actions=[
                            codepipeline.Actions(
                                Name='Build',
                                ActionTypeId=codepipeline.ActionTypeId(
                                    Category='Build',
                                    Owner='AWS',
                                    Provider='CodeBuild',
                                    Version='1'
                                ),
                                Configuration={
                                    'ProjectName': buildproject.ref()
                                },
                                InputArtifacts=[
                                    codepipeline.InputArtifacts(
                                        Name='CodeCommitRepo'
                                    )
                                ]
                            )
                        ]
                    )
                ]
            )
        )


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    from stacker.context import Context
    print(Pipeline('test', Context({"namespace": "test"}), None).to_json())



"""CFNgin diff action."""
import logging
from operator import attrgetter

from .. import exceptions
from ..status import (COMPLETE, INTERRUPTED, NotSubmittedStatus,
                      NotUpdatedStatus, SkippedStatus)
from ..status import StackDoesNotExist as StackDoesNotExistStatus
from . import build
from .base import build_walker

LOGGER = logging.getLogger(__name__)


class DictValue(object):
    """Used to create a diff of two dictionaries."""

    ADDED = "ADDED"
    REMOVED = "REMOVED"
    MODIFIED = "MODIFIED"
    UNMODIFIED = "UNMODIFIED"

    formatter = "%s%s = %s"

    def __init__(self, key, old_value, new_value):
        """Instantiate class."""
        self.key = key
        self.old_value = old_value
        self.new_value = new_value

    def __eq__(self, other):
        """Compare if self is equal to another object."""
        return self.__dict__ == other.__dict__

    def changes(self):
        """Return changes to represent the diff between old and new value.

        Returns:
            List[str]: Representation of the change (if any) between old and
            new value.

        """
        output = []
        if self.status() is self.UNMODIFIED:
            output = [self.formatter % (' ', self.key, self.old_value)]
        elif self.status() is self.ADDED:
            output.append(self.formatter % ('+', self.key, self.new_value))
        elif self.status() is self.REMOVED:
            output.append(self.formatter % ('-', self.key, self.old_value))
        elif self.status() is self.MODIFIED:
            output.append(self.formatter % ('-', self.key, self.old_value))
            output.append(self.formatter % ('+', self.key, self.new_value))
        return output

    def status(self):
        """Status of changes between the old value and new value."""
        if self.old_value == self.new_value:
            return self.UNMODIFIED
        if self.old_value is None:
            return self.ADDED
        if self.new_value is None:
            return self.REMOVED
        return self.MODIFIED


def diff_dictionaries(old_dict, new_dict):
    """Calculate the diff two single dimension dictionaries.

    Args:
        old_dict(Dict[Any, Any]): Old dictionary.
        new_dict(Dict[Any, Any]): New dictionary.

    Returns:
        Tuple[int, List[:class:`DictValue`]]: Number of changed records and
        the :class:`DictValue` object containing the changes.

    """
    old_set = set(old_dict)
    new_set = set(new_dict)

    added_set = new_set - old_set
    removed_set = old_set - new_set
    common_set = old_set & new_set

    changes = 0
    output = []
    for key in added_set:
        changes += 1
        output.append(DictValue(key, None, new_dict[key]))

    for key in removed_set:
        changes += 1
        output.append(DictValue(key, old_dict[key], None))

    for key in common_set:
        output.append(DictValue(key, old_dict[key], new_dict[key]))
        if str(old_dict[key]) != str(new_dict[key]):
            changes += 1

    output.sort(key=attrgetter("key"))
    return changes, output


def format_params_diff(parameter_diff):
    """Handle the formatting of differences in parameters.

    Args:
        parameter_diff (List[:class:`DictValue`]): A list of
            :class:`DictValue` detailing the differences between two dicts
            returned by :func:`diff_dictionaries`.

    Returns:
        str: A formatted string that represents a parameter diff

    """
    params_output = '\n'.join([line for v in parameter_diff
                               for line in v.changes()])
    return """--- Old Parameters
+++ New Parameters
******************
%s\n""" % params_output


def diff_parameters(old_params, new_params):
    """Compare the old vs. new parameters and returns a "diff".

    If there are no changes, we return an empty list.

    Args:
        old_params(Dict[Any, Any]): old paramters
        new_params(Dict[Any, Any]): new parameters

    Returns:
        List[:class:`DictValue`]: A list of differences.

    """
    changes, diff = diff_dictionaries(old_params, new_params)
    if changes == 0:
        return []
    return diff


class Action(build.Action):
    """Responsible for diffing CloudFormation stacks in AWS and locally.

    Generates the build plan based on stack dependencies (these dependencies
    are determined automatically based on references to output values from
    other stacks).

    The plan is then used to create a changeset for a stack using a
    generated template based on the current config.

    """

    DESCRIPTION = 'Diff stacks'

    @property
    def _stack_action(self):
        """Run against a step."""
        return self._diff_stack

    def _diff_stack(self, stack, **_kwargs):  # pylint: disable=too-many-return-statements
        """Handle diffing a stack in CloudFormation vs our config."""
        if self.cancel.wait(0):
            return INTERRUPTED

        if not build.should_submit(stack):
            return NotSubmittedStatus()

        if not build.should_update(stack):
            return NotUpdatedStatus()

        provider = self.build_provider(stack)
        tags = build.build_stack_tags(stack)

        stack.resolve(self.context, provider)
        parameters = self.build_parameters(stack)

        try:
            outputs = provider.get_stack_changes(
                stack, self._template(stack.blueprint), parameters, tags
            )
            stack.set_outputs(outputs)
        except exceptions.StackDidNotChange:
            LOGGER.info('No changes: %s', stack.fqn)
        except exceptions.StackDoesNotExist:  # TODO make sure this is correct post chageset
            if self.context.persistent_graph:
                return SkippedStatus('persistent graph: stack does not '
                                     'exist, will be removed')
            return StackDoesNotExistStatus()
        except AttributeError as err:  # TODO make sure this is correct post chageset
            if (self.context.persistent_graph and
                    'defined class or template path' in str(err)):
                return SkippedStatus('persistent graph: will be destroyed')
            raise
        return COMPLETE

    def run(self, **kwargs):
        """Kicks off the diffing of the stacks in the stack_definitions."""
        plan = self._generate_plan(require_unlocked=False,
                                   include_persistent_graph=True)
        plan.outline(logging.DEBUG)
        if plan.keys():
            LOGGER.info("Diffing stacks: %s", ", ".join(plan.keys()))
        else:
            LOGGER.warning('WARNING: No stacks detected (error in config?)')
        walker = build_walker(kwargs.get('concurrency', 0))
        plan.execute(walker)

    def pre_run(self, **kwargs):
        """Do nothing."""

    def post_run(self, **kwargs):
        """Do nothing."""



"""CFNgin test."""
# flake8: noqa
# pylint: disable=invalid-name
from os.path import basename

from integration_tests.test_cfngin.test_cfngin import Cfngin

FILE_BASENAME = '.'.join(basename(__file__).split('.')[:-1])


class DestroyRemoved(Cfngin):
    """Test CFNgin persistent graph destroying removed stacks.

    Requires valid AWS credentials.

    """

    REQUIRED_FIXTURE_FILES = [FILE_BASENAME + '.1.yaml',
                              FILE_BASENAME + '.2.yaml']
    TEST_NAME = __name__

    def _build(self):
        """Execute and assert initial build."""
        code, _stdout, stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'
        expected_lines = [
            'other was removed from the Stacker config file so it is being destroyed.',
            'other: submitted (submitted for destruction)',
            'other: complete (stack destroyed)'
        ]
        for stack in ['vpc', 'bastion', 'other']:
            expected_lines.append(f'{stack}: submitted (creating new stack)')
            expected_lines.append(f'{stack}: complete (creating new stack)')
            if stack != 'other':
                expected_lines.append(f'{stack}: skipped (nochange)')
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def run(self):
        """Run the test."""
        self.copy_fixtures()
        self._build()

    def teardown(self):
        """Teardown any created resources and delete files."""
        self.runway_cmd('destroy')
        self.cleanup_fixtures()



""".. Base class for lookup handlers.

Lookup Arguments
^^^^^^^^^^^^^^^^

.. _lookup arguments:

Arguments can be passed to Lookups to effect how they function.

To provide arguments to a Lookup, use a double-colon (``::``) after the
query. Each argument is then defined as a **key** and **value** seperated with
equals (``=``) and the arguments theselves are seperated with a comma (``,``).
The arguments can have an optional space after the comma and before the next
key to make them easier to read but this is not required. The value of all
arguments are read as strings.

.. rubric:: Example
.. code-block:: yaml

    ${var my_query::default=true, transform=bool}
    ${env MY_QUERY::default=1,transform=bool}

Each Lookup may have their own, specific arguments that it uses to modify its
functionality or the value it returns. There is also a common set of arguments
that all Lookups accept.

Common Arguments
~~~~~~~~~~~~~~~~

.. _common lookup arguments:

+---------------+-------------------------------------------------------------+
| Argument      | Description                                                 |
+===============+=============================================================+
| ``default``   | If the Lookup is unable to find a value for the provided    |
|               | query, this value will be returned instead of raising a     |
|               | ``ValueError``.                                             |
+---------------+-------------------------------------------------------------+
| ``transform`` | Transform the data returned by a Lookup into a different    |
|               | datatype. Supports ``str`` and ``bool``.                    |
+---------------+-------------------------------------------------------------+

.. rubric:: Example
.. code-block:: yaml

  deployments:
    - environments:
        some_variable: ${var some_value::default=my_value}
        comma_list: ${var my_list::default=undefined, transform=str}

"""
# python2 supported pylint is unable to load this when in a venv
from distutils.util import strtobool  # pylint: disable=E
from typing import (TYPE_CHECKING, Any, Dict,  # noqa: F401 pylint: disable=W
                    Tuple, Union)

from six import string_types

from ...cfngin.util import read_value_from_path

# python2 supported pylint sees this is cyclic even though its only for type checking
# pylint: disable=cyclic-import
if TYPE_CHECKING:
    from ...context import Context  # noqa: F401 pylint: disable=unused-import


class LookupHandler(object):
    """Base class for lookup handlers."""

    @classmethod
    def dependencies(cls, _lookup_data):
        """Calculate any dependencies required to perform this lookup.

        Note that lookup_data may not be (completely) resolved at this time.

        Args:
            lookup_data (VariableValue): Parameter(s) given to this lookup.

        Returns:
            Set

        """
        return set()

    @classmethod
    def handle(cls, value, context, **kwargs):
        # type: (str, 'Context', Any) -> Any
        """Perform the lookup.

        Args:
            value: Parameter(s) given to the lookup.
            context: The current context object.
            provider: Optional provider to use when handling the lookup.

        Returns:
            (Any) Looked-up value.

        """
        raise NotImplementedError

    @classmethod
    def parse(cls, value):
        # type: (str) -> Tuple[str, Dict[str, str]]
        """Parse the value passed to a lookup in a standardized way.

        Args:
            value: The raw value passed to a lookup.

        Returns:
            The lookup query and a dict of arguments

        """
        raw_value = read_value_from_path(value)

        colon_split = raw_value.split('::', 1)

        query = colon_split.pop(0)
        args = cls._parse_args(colon_split[0]) if colon_split else {}

        return query, args

    @classmethod
    def _parse_args(cls, args):
        # type: (str) -> Dict[str, str]
        """Convert a string into an args dict.

        Each arg should be seporated by  ``,``. The key and value should
        be seporated by ``=``. Any leading or following spaces are stripped.

        Args:
            args: A string containing arguments to be parsed. (e.g.
                ``'key1=value1, key2=value2'``)

        Returns:
            Dict of parsed args.

        """
        split_args = args.split(',')
        return {key.strip(): value.strip() for key, value in
                [arg.split('=') for arg in split_args]}

    @classmethod
    def transform(cls, value, to_type='str', **kwargs):
        # type: (str, str, Any) -> Any
        """Transform the result of a lookup into another datatype.

        Args:
            value: What is to be transformed.
            to_type: The type the value will be transformed into.

        Returns:
            The transformed value.

        """
        mapping = {
            'bool': cls._transform_to_bool,
            'str': cls._transform_to_string
        }

        if not to_type:
            return value

        return mapping[to_type](value, **kwargs)  # type: ignore

    @classmethod
    def _transform_to_bool(cls, value, **_):
        # type: (Union[bool, str], Any) -> bool
        """Transform a string into a bool.

        Args:
            value: The value to be transformed into a bool.

        Raises:
            ValueError: The value provided was not a bool or string or
                the string could not be converted to a bool.

        """
        if isinstance(value, bool):
            return value
        if isinstance(value, string_types):
            return bool(strtobool(value))
        raise TypeError('Value must be a string or bool to use transform=bool. '
                        'Got type {}.'.format(type(value)))

    @classmethod
    def _transform_to_string(cls, value, delimiter=None, **_):
        # type: (Any, str, Any) -> str
        """Transform anything into a string.

        If the datatype of ``value`` is a list or similar to a list, ``join()``
        is used to construct the list using a given delimiter or ``,``.

        Args:
            value: The value to be transformed into a string.
            delimited: Used when transforming a list like object into a string
                to join each element together.

        """
        if isinstance(value, (list, set, tuple)):
            return '{}'.format(delimiter or ',').join(value)
        return str(value)



"""Test getting current user."""
import os
import json

from subprocess import check_output
from integration_tests.test_commands.test_commands import Commands


class TestRunAWS(Commands):
    """Tests run-aws subcommand."""

    TEST_NAME = __name__

    def init(self):
        """Initialize test."""
        pass  # pylint: disable=unnecessary-pass

    def run(self):
        """Run test."""
        response = check_output(
            ['runway',
             'run-aws',
             'sts',
             'get-caller-identity']
        ).decode()
        data = json.loads(response)
        assert 'Arn' in data, 'response has no Arn property'

    def teardown(self):
        """Teardown any created resources."""
        pass  # pylint: disable=unnecessary-pass



"""AMI lookup."""
# pylint: disable=unused-argument,line-too-long,arguments-differ
import operator
import re

from runway.lookups.handlers.base import LookupHandler

from ...session_cache import get_session
from ...util import read_value_from_path

TYPE_NAME = "ami"


class ImageNotFound(Exception):
    """Image not found."""

    def __init__(self, search_string):
        """Instantiate class."""
        self.search_string = search_string
        message = ("Unable to find ec2 image with search string: {}").format(
            search_string
        )
        super(ImageNotFound, self).__init__(message)


class AmiLookup(LookupHandler):
    """AMI lookup."""

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        """Fetch the most recent AMI Id using a filter.

        Args:
            value (str): Parameter(s) given to this lookup.
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        Returns:
            str: Looked up value.

        Example:
            .. code-block:

                ${ami [<region>@]owners:self,account,amazon name_regex:serverX-[0-9]+ architecture:x64,i386}

            The above fetches the most recent AMI where owner is self
            account or amazon and the ami name matches the regex described,
            the architecture will be either x64 or i386

            You can also optionally specify the region in which to perform the
            AMI lookup.

            Valid arguments:

            owners (comma delimited) REQUIRED ONCE:
                aws_account_id | amazon | self

            name_regex (a regex) REQUIRED ONCE:
                e.g. ``my-ubuntu-server-[0-9]+``

            executable_users (comma delimited) OPTIONAL ONCE:
                aws_account_id | amazon | self

            Any other arguments specified are sent as filters to the aws api
            For example, ``architecture:x86_64`` will add a filter

        """  # noqa
        value = read_value_from_path(value)

        if "@" in value:
            region, value = value.split("@", 1)
        else:
            region = provider.region

        ec2 = get_session(region).client('ec2')

        values = {}
        describe_args = {}

        # now find any other arguments that can be filters
        matches = re.findall(r'([0-9a-zA-z_-]+:[^\s$]+)', value)
        for match in matches:
            k, v = match.split(':', 1)
            values[k] = v

        if not values.get('owners'):
            raise Exception("'owners' value required when using ami")
        owners = values.pop('owners').split(',')
        describe_args["Owners"] = owners

        if not values.get('name_regex'):
            raise Exception("'name_regex' value required when using ami")
        name_regex = values.pop('name_regex')

        if values.get('executable_users'):
            executable_users = values.pop('executable_users').split(',')
            describe_args["ExecutableUsers"] = executable_users

        filters = []
        for k, v in values.items():
            filters.append({"Name": k, "Values": v.split(',')})
        describe_args["Filters"] = filters

        result = ec2.describe_images(**describe_args)

        images = sorted(result['Images'],
                        key=operator.itemgetter('CreationDate'),
                        reverse=True)
        for image in images:
            # sometimes we get ARI/AKI in response - these don't have a 'Name'
            if re.match("^%s$" % name_regex, image.get('Name', '')):
                return image['ImageId']

        raise ImageNotFound(value)



"""Mock hook."""
# pylint: disable=unused-argument


def mock_hook(provider, context, **kwargs):
    """Mock hook.

    Returns:
        {'result': kwargs['value']}

    """
    return {'result': kwargs['value']}



"""Tests for runway.cfngin.blueprints.raw."""
import json
import unittest

from mock import MagicMock

from runway.cfngin.blueprints.raw import (RawTemplateBlueprint,
                                          get_template_params,
                                          get_template_path)
from runway.variables import Variable

from ..factories import mock_context

RAW_JSON_TEMPLATE_PATH = 'tests/cfngin/fixtures/cfn_template.json'
RAW_YAML_TEMPLATE_PATH = 'tests/cfngin/fixtures/cfn_template.yaml'
RAW_J2_TEMPLATE_PATH = 'tests/cfngin/fixtures/cfn_template.json.j2'


def test_get_template_path_local_file(tmpdir):
    """Verify get_template_path finding a file relative to CWD."""
    template_path = tmpdir.join('cfn_template.json')
    template_path.ensure()

    with tmpdir.as_cwd():
        result = get_template_path('cfn_template.json')
        assert template_path.samefile(result)


def test_get_template_path_invalid_file(tmpdir):
    """Verify get_template_path with an invalid filename."""
    with tmpdir.as_cwd():
        assert get_template_path('cfn_template.json') is None


def test_get_template_path_file_in_syspath(tmpdir, monkeypatch):
    """Verify get_template_path with a file in sys.path.

    This ensures templates are able to be retrieved from remote packages.

    """
    template_path = tmpdir.join('cfn_template.json')
    template_path.ensure()

    monkeypatch.syspath_prepend(tmpdir)
    result = get_template_path(template_path.basename)
    assert template_path.samefile(result)


def test_get_template_params():
    """Verify get_template_params function operation."""
    template_dict = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "TestTemplate",
        "Parameters": {
            "Param1": {
                "Type": "String"
            },
            "Param2": {
                "Default": "default",
                "Type": "CommaDelimitedList"
            }
        },
        "Resources": {}
    }
    template_params = {
        "Param1": {
            "Type": "String"
        },
        "Param2": {
            "Default": "default",
            "Type": "CommaDelimitedList"
        }
    }

    assert get_template_params(template_dict) == template_params


class TestBlueprintRendering(unittest.TestCase):
    """Test class for blueprint rendering."""

    def test_to_json(self):
        """Verify to_json method operation."""
        expected_json = json.dumps(
            {
                "AWSTemplateFormatVersion": "2010-09-09",
                "Description": "TestTemplate",
                "Parameters": {
                    "Param1": {
                        "Type": "String"
                    },
                    "Param2": {
                        "Default": "default",
                        "Type": "CommaDelimitedList"
                    }
                },
                "Resources": {
                    "Dummy": {
                        "Type": "AWS::CloudFormation::WaitConditionHandle"
                    }
                },
                "Outputs": {
                    "DummyId": {
                        "Value": "dummy-1234"
                    }
                }
            },
            sort_keys=True,
            indent=4
        )
        self.assertEqual(
            RawTemplateBlueprint(
                name="test",
                context=mock_context(),
                raw_template_path=RAW_JSON_TEMPLATE_PATH).to_json(),
            expected_json
        )

    def test_j2_to_json(self):
        """Verify jinja2 template parsing."""
        expected_json = json.dumps(
            {
                "AWSTemplateFormatVersion": "2010-09-09",
                "Description": "TestTemplate",
                "Parameters": {
                    "Param1": {
                        "Type": "String"
                    },
                    "Param2": {
                        "Default": "default",
                        "Type": "CommaDelimitedList"
                    }
                },
                "Resources": {
                    "Dummy": {
                        "Type": "AWS::CloudFormation::WaitConditionHandle"
                    }
                },
                "Outputs": {
                    "DummyId": {
                        "Value": "dummy-bar-param1val-foo-1234"
                    }
                }
            },
            sort_keys=True,
            indent=4
        )
        blueprint = RawTemplateBlueprint(
            name="stack1",
            context=mock_context(
                extra_config_args={'stacks': [{'name': 'stack1',
                                               'template_path': 'unused',
                                               'variables': {
                                                   'Param1': 'param1val',
                                                   'bar': 'foo'}}]},
                environment={'foo': 'bar'}),
            raw_template_path=RAW_J2_TEMPLATE_PATH
        )
        blueprint.resolve_variables([Variable("Param1", "param1val", 'cfngin'),
                                     Variable("bar", "foo", 'cfngin')])
        self.assertEqual(
            expected_json,
            blueprint.to_json()
        )


class TestVariables(unittest.TestCase):
    """Test class for blueprint variable methods."""

    def test_get_parameter_definitions_json(self):  # noqa pylint: disable=invalid-name
        """Verify get_parameter_definitions method with json raw template."""
        blueprint = RawTemplateBlueprint(
            name="test",
            context=MagicMock(),
            raw_template_path=RAW_JSON_TEMPLATE_PATH)
        parameters = blueprint.get_parameter_definitions()
        self.assertEqual(
            parameters,
            {"Param1": {"Type": "String"},
             "Param2": {"Default": "default",
                        "Type": "CommaDelimitedList"}})

    def test_get_parameter_definitions_yaml(self):  # noqa pylint: disable=invalid-name
        """Verify get_parameter_definitions method with yaml raw template."""
        blueprint = RawTemplateBlueprint(
            name="test",
            context=MagicMock(),
            raw_template_path=RAW_YAML_TEMPLATE_PATH
        )
        parameters = blueprint.get_parameter_definitions()
        self.assertEqual(
            parameters,
            {"Param1": {"Type": "String"},
             "Param2": {"Default": "default",
                        "Type": "CommaDelimitedList"}})

    def test_get_required_parameter_definitions_json(self):  # noqa pylint: disable=invalid-name
        """Verify get_required_param... method with json raw template."""
        blueprint = RawTemplateBlueprint(
            name="test",
            context=MagicMock(),
            raw_template_path=RAW_JSON_TEMPLATE_PATH
        )
        self.assertEqual(
            blueprint.get_required_parameter_definitions(),
            {"Param1": {"Type": "String"}})

    def test_get_required_parameter_definitions_yaml(self):  # noqa pylint: disable=invalid-name
        """Verify get_required_param... method with yaml raw template."""
        blueprint = RawTemplateBlueprint(
            name="test",
            context=MagicMock(),
            raw_template_path=RAW_YAML_TEMPLATE_PATH
        )
        self.assertEqual(
            blueprint.get_required_parameter_definitions(),
            {"Param1": {"Type": "String"}})



"""KMS translator.

.. important:: The translator is going to be deprecated in favor of the lookup.

"""
from ...lookups.handlers.kms import KmsLookup


def kms_simple_constructor(loader, node):
    """KMS simple constructor."""
    value = loader.construct_scalar(node)
    return KmsLookup.handle(value)



"""Testing that module type detection is handled appropriately."""
import os
import glob

from send2trash import send2trash

from integration_tests.integration_test import IntegrationTest
from integration_tests.util import (copy_file, copy_dir, import_tests,
                                    execute_tests)


class RunwayModuleTypeDetection(IntegrationTest):
    """Test that module type detection is handled."""

    base_dir = os.path.abspath(os.path.dirname(__file__))
    fixtures_dir = os.path.join(base_dir, 'fixtures')
    tests_dir = os.path.join(base_dir, 'tests')

    mtd_test_dir = os.path.join(base_dir, 'runway_module_type_detection_test_dir')

    def copy_fixture(self, name='two-regions-app.cfn'):
        """Copy fixture files for test."""
        copy_dir(
            os.path.join(self.fixtures_dir, name),
            os.path.join(self.mtd_test_dir, name)
        )

    def copy_runway(self, template):
        """Copy runway template to proper directory."""
        template_file = os.path.join(self.fixtures_dir, 'runway-{}.yml'.format(template))
        copy_file(template_file, os.path.join(self.mtd_test_dir, 'runway.yml'))

    def run(self):
        """Find all tests and run them."""
        import_tests(self.logger, self.tests_dir, 'test_*')
        tests = [
            test(self.logger) for test in RunwayModuleTypeDetection.__subclasses__()
        ]
        if not tests:
            raise Exception('No tests were found.')
        self.logger.debug('FOUND TESTS: %s', tests)
        self.set_environment('dev')
        err_count = execute_tests(tests, self.logger)
        assert err_count == 0
        return err_count

    def clean(self):
        """Clean up fixture test directory."""
        file_types = ('*.yaml', '*.yml')
        templates = []
        for file_type in file_types:
            templates.extend(glob.glob(os.path.join(
                self.mtd_test_dir,
                file_type
            )))
        for template in templates:
            if os.path.isfile(template):
                self.logger.debug('send2trash: "%s"', template)
                send2trash(template)
        folders = ['']
        for folder in folders:
            folder_path = os.path.join(self.mtd_test_dir, folder)
            if os.path.isdir(folder_path):
                self.logger.debug('send2trash: "%s"', folder_path)
                send2trash(folder_path)

    def teardown(self):
        """Teardown resources create during init."""
        self.clean()



"""AWS ECS hook.

A lot of this code exists to deal w/ the broken ECS connect_to_region
function, and will be removed once this pull request is accepted:
https://github.com/boto/boto/pull/3143

"""
# pylint: disable=unused-argument
import logging

from six import string_types

from ..session_cache import get_session

LOGGER = logging.getLogger(__name__)


def create_clusters(provider, context, **kwargs):
    """Create ECS clusters.

    Args:
        provider (:class:`runway.cfngin.providers.base.BaseProvider`): Provider
            instance. (passed in by CFNgin)
        context (:class:`runway.cfngin.context.Context`): Context instance.
            (passed in by CFNgin)

    Keyword Args:
        clusters (List[str]): Names of clusters to create.


    Returns:
        bool: Whether or not the hook succeeded.

    """
    conn = get_session(provider.region).client('ecs')

    try:
        clusters = kwargs["clusters"]
    except KeyError:
        LOGGER.error("setup_clusters hook missing \"clusters\" argument")
        return False

    if isinstance(clusters, string_types):
        clusters = [clusters]

    cluster_info = {}
    for cluster in clusters:
        LOGGER.debug("Creating ECS cluster: %s", cluster)
        response = conn.create_cluster(clusterName=cluster)
        cluster_info[response["cluster"]["clusterName"]] = response
    return {"clusters": cluster_info}



"""CFNgin graph action."""
import json
import logging
import sys

from ..plan import merge_graphs
from .base import BaseAction

LOGGER = logging.getLogger(__name__)


def each_step(graph):
    """Yield each step and it's direct dependencies.

    Args:
        graph (:class:`runway.cfngin.plan.Graph`): Graph to iterate over.

    Yields:
        Tuple[Step, Set(str)]

    """
    steps = graph.topological_sort()
    steps.reverse()

    for step in steps:
        deps = graph.downstream(step.name)
        yield step, deps


def dot_format(out, graph, name="digraph"):
    """Output a graph using the graphviz "dot" format.

    Args:
        out (TextIo): Where output will be written.
        graph (:class:`runway.cfngin.plan.Graph`): Graph to be output.
        name (str): Name of the graph.

    """
    out.write("digraph %s {\n" % name)
    for step, deps in each_step(graph):
        for dep in deps:
            out.write("  \"%s\" -> \"%s\";\n" % (step, dep))

    out.write("}\n")


def json_format(out, graph):
    """Output the graph in a machine readable JSON format.

    Args:
        out (TextIo): Where output will be written.
        graph (:class:`runway.cfngin.plan.Graph`): Graph to be output.

    """
    steps = {}
    for step, deps in each_step(graph):
        steps[step.name] = {}
        steps[step.name]["deps"] = [dep.name for dep in deps]

    json.dump({"steps": steps}, out, indent=4)
    out.write("\n")


FORMATTERS = {
    "dot": dot_format,
    "json": json_format,
}


class Action(BaseAction):
    """Responsible for outputing a graph for the current CFNgin config."""

    DESCRIPTION = 'Print graph'

    @property
    def _stack_action(self):
        """Run against a step."""
        return None

    def run(self, **kwargs):
        """Generate the underlying graph and prints it."""
        graph = self._generate_plan(require_unlocked=False,
                                    include_persistent_graph=True).graph
        if self.context.persistent_graph:
            graph = merge_graphs(self.context.persistent_graph, graph)
        if kwargs.get('reduce'):
            # This will performa a transitive reduction on the underlying
            # graph, producing less edges. Mostly useful for the "dot" format,
            # when converting to PNG, so it creates a prettier/cleaner
            # dependency graph.
            graph.transitive_reduction()

        fn = FORMATTERS[kwargs.get('format')]
        fn(sys.stdout, graph)
        sys.stdout.flush()



"""Tests for runway.cfngin.lookups.handlers.output."""
import unittest

from mock import MagicMock

from runway.cfngin.lookups.handlers.output import OutputLookup
from runway.cfngin.stack import Stack

from ...factories import generate_definition


class TestOutputHandler(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.output.OutputLookup."""

    def setUp(self):
        """Run before tests."""
        self.context = MagicMock()

    def test_output_handler(self):
        """Test output handler."""
        stack = Stack(
            definition=generate_definition("vpc", 1),
            context=self.context)
        stack.set_outputs({
            "SomeOutput": "Test Output"})
        self.context.get_stack.return_value = stack
        value = OutputLookup.handle("stack-name::SomeOutput",
                                    context=self.context)
        self.assertEqual(value, "Test Output")
        self.assertEqual(self.context.get_stack.call_count, 1)
        args = self.context.get_stack.call_args
        self.assertEqual(args[0][0], "stack-name")



"""Integration tests for runway.cfngin."""
from integration_tests.integration_test import IntegrationTest
from integration_tests.util import execute_tests, import_tests


class Cfngin(IntegrationTest):
    """Test runner for CFNgin."""

    def run(self):
        """Find all tests and run them."""
        import_tests(self.logger, self.tests_dir, '[0-9][0-9]_**')
        self.set_environment('dev')
        self.set_env_var('CI', '1')
        tests = [test(self.logger, self.environment)
                 for test in Cfngin.__subclasses__()]
        if not tests:
            raise Exception('No tests were found.')
        self.logger.debug('FOUND TESTS: %s', tests)
        err_count = execute_tests(tests, self.logger)
        assert err_count == 0  # assert that all subtests were successful
        return err_count

    def teardown(self):
        """Teardown resources create during init."""
        self.unset_env_var('CI')
        self.logger.debug('Teardown is defined in the submodules, not '
                          'the "TestCfngin" parent class.')



# encoding: utf-8
"""Tests for runway.cfngin.lookups.handlers.file.

.. note: ``encoding: utf-8`` is required for python2 support due to a character
         in a string in ``test_yaml_codec_raw``.

"""
import base64
import json
import unittest

import mock
import yaml
from troposphere import Base64, GenericHelperFn, Join

from runway.cfngin.lookups.handlers.file import (FileLookup, json_codec,
                                                 parameterized_codec,
                                                 yaml_codec)


def to_template_dict(obj):
    """Extract the CFN template dict of an object for test comparisons."""
    if hasattr(obj, 'to_dict') and callable(obj.to_dict):
        return obj.to_dict()
    elif isinstance(obj, dict):
        return dict((key, to_template_dict(value))
                    for (key, value) in obj.items())
    elif isinstance(obj, (list, tuple)):
        return type(obj)(to_template_dict(item) for item in obj)
    else:
        return obj


class TestFileTranslator(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.file.FileLookup."""

    @staticmethod
    def assertTemplateEqual(left, right):  # noqa: N802 pylint: disable=invalid-name
        """Assert that two codec results are equivalent.

        Convert both sides to their template representations, since Troposphere
        objects are not natively comparable.

        """
        return to_template_dict(left) == to_template_dict(right)

    def test_parameterized_codec_b64(self):
        """Test parameterized codec b64."""
        expected = Base64(
            Join(u'', [u'Test ', {u'Ref': u'Interpolation'}, u' Here'])
        )

        out = parameterized_codec(u'Test {{Interpolation}} Here', True)
        self.assertEqual(Base64, out.__class__)
        self.assertTemplateEqual(expected, out)

    def test_parameterized_codec_plain(self):
        """Test parameterized codec plain."""
        expected = Join(u'', [u'Test ', {u'Ref': u'Interpolation'}, u' Here'])

        out = parameterized_codec(u'Test {{Interpolation}} Here', False)
        self.assertEqual(GenericHelperFn, out.__class__)
        self.assertTemplateEqual(expected, out)

    def test_parameterized_codec_plain_no_interpolation(self):
        """Test parameterized codec plain no interpolation."""
        expected = u'Test Without Interpolation Here'

        out = parameterized_codec(u'Test Without Interpolation Here', False)
        self.assertEqual(GenericHelperFn, out.__class__)
        self.assertTemplateEqual(expected, out)

    def test_yaml_codec_raw(self):
        """Test yaml codec raw."""
        structured = {
            u'Test': [1, None, u'unicode ✓', {u'some': u'obj'}]
        }
        # Note: must use safe_dump, since regular dump adds !python/unicode
        # tags, which we don't want, or we can't be sure we're correctly
        # loading string as unicode.
        raw = yaml.safe_dump(structured)

        out = yaml_codec(raw, parameterized=False)
        self.assertEqual(structured, out)

    def test_yaml_codec_parameterized(self):
        """Test yaml codec parameterized."""
        processed = {
            u'Test': Join(u'', [u'Test ', {u'Ref': u'Interpolation'},
                                u' Here'])
        }
        structured = {
            u'Test': u'Test {{Interpolation}} Here'
        }
        raw = yaml.safe_dump(structured)

        out = yaml_codec(raw, parameterized=True)
        self.assertTemplateEqual(processed, out)

    def test_json_codec_raw(self):
        """Test json codec raw."""
        structured = {
            u'Test': [1, None, u'str', {u'some': u'obj'}]
        }
        raw = json.dumps(structured)

        out = json_codec(raw, parameterized=False)
        self.assertEqual(structured, out)

    def test_json_codec_parameterized(self):
        """Test json codec parameterized."""
        processed = {
            u'Test': Join(u'', [u'Test ', {u'Ref': u'Interpolation'},
                                u' Here'])
        }
        structured = {
            u'Test': u'Test {{Interpolation}} Here'
        }
        raw = json.dumps(structured)

        out = json_codec(raw, parameterized=True)
        self.assertTemplateEqual(processed, out)

    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path',
                return_value='')
    def test_file_loaded(self, content_mock):
        """Test file loaded."""
        FileLookup.handle(u'plain:file://tmp/test')
        content_mock.assert_called_with(u'file://tmp/test')

    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path',
                return_value=u'Hello, world')
    def test_handler_plain(self, _):
        """Test handler plain."""
        out = FileLookup.handle(u'plain:file://tmp/test')
        self.assertEqual(u'Hello, world', out)

    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path')
    def test_handler_b64(self, content_mock):
        """Test handler b64."""
        plain = u'Hello, world'
        encoded = base64.b64encode(plain.encode('utf8')).decode('utf-8')

        content_mock.return_value = plain
        out = FileLookup.handle(u'base64:file://tmp/test')
        self.assertEqual(encoded, out)

    @mock.patch('runway.cfngin.lookups.handlers.file.parameterized_codec')
    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path')
    def test_handler_parameterized(self, content_mock, codec_mock):
        """Test handler parameterized."""
        result = mock.Mock()
        codec_mock.return_value = result

        out = FileLookup.handle(u'parameterized:file://tmp/test')
        codec_mock.assert_called_once_with(content_mock.return_value, False)

        self.assertEqual(result, out)

    @mock.patch('runway.cfngin.lookups.handlers.file.parameterized_codec')
    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path')
    def test_handler_parameterized_b64(self, content_mock, codec_mock):
        """Test handler parameterized b64."""
        result = mock.Mock()
        codec_mock.return_value = result

        out = FileLookup.handle(u'parameterized-b64:file://tmp/test')
        codec_mock.assert_called_once_with(content_mock.return_value, True)

        self.assertEqual(result, out)

    @mock.patch('runway.cfngin.lookups.handlers.file.yaml_codec')
    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path')
    def test_handler_yaml(self, content_mock, codec_mock):
        """Test handler yaml."""
        result = mock.Mock()
        codec_mock.return_value = result

        out = FileLookup.handle(u'yaml:file://tmp/test')
        codec_mock.assert_called_once_with(content_mock.return_value,
                                           parameterized=False)

        self.assertEqual(result, out)

    @mock.patch('runway.cfngin.lookups.handlers.file.yaml_codec')
    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path')
    def test_handler_yaml_parameterized(self, content_mock, codec_mock):
        """Test handler yaml parameterized."""
        result = mock.Mock()
        codec_mock.return_value = result

        out = FileLookup.handle(u'yaml-parameterized:file://tmp/test')
        codec_mock.assert_called_once_with(content_mock.return_value,
                                           parameterized=True)

        self.assertEqual(result, out)

    @mock.patch('runway.cfngin.lookups.handlers.file.json_codec')
    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path')
    def test_handler_json(self, content_mock, codec_mock):
        """Test handler json."""
        result = mock.Mock()
        codec_mock.return_value = result

        out = FileLookup.handle(u'json:file://tmp/test')
        codec_mock.assert_called_once_with(content_mock.return_value,
                                           parameterized=False)

        self.assertEqual(result, out)

    @mock.patch('runway.cfngin.lookups.handlers.file.json_codec')
    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path')
    def test_handler_json_parameterized(self, content_mock, codec_mock):
        """Test handler json parameterized."""
        result = mock.Mock()
        codec_mock.return_value = result

        out = FileLookup.handle(u'json-parameterized:file://tmp/test')
        codec_mock.assert_called_once_with(content_mock.return_value,
                                           parameterized=True)

        self.assertEqual(result, out)

    @mock.patch('runway.cfngin.lookups.handlers.file.read_value_from_path')
    def test_unknown_codec(self, _):
        """Test unknown codec."""
        with self.assertRaises(KeyError):
            FileLookup.handle(u'bad:file://tmp/test')



"""AWS KMS lookup."""
# pylint: disable=arguments-differ,unused-argument
import codecs

from runway.lookups.handlers.base import LookupHandler

from ...session_cache import get_session
from ...util import read_value_from_path

TYPE_NAME = "kms"


class KmsLookup(LookupHandler):
    """AWS KMS lookup."""

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        r"""Decrypt the specified value with a master key in KMS.

        Args:
            value (str): Parameter(s) given to this lookup.
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        ``value`` should be in the following format:

            [<region>@]<base64 encrypted value>

        .. note: The region is optional, and defaults to the environment's
                 ``AWS_DEFAULT_REGION`` if not specified.

        Example:
            ::

                # We use the aws cli to get the encrypted value for the string
                # "PASSWORD" using the master key called "myKey" in
                # us-east-1
                $ aws --region us-east-1 kms encrypt --key-id alias/myKey \
                        --plaintext "PASSWORD" --output text --query CiphertextBlob

                CiD6bC8t2Y<...encrypted blob...>

                # With CFNgin we would reference the encrypted value like:
                conf_key: ${kms us-east-1@CiD6bC8t2Y<...encrypted blob...>}

            You can optionally store the encrypted value in a file, ie::

                kms_value.txt
                us-east-1@CiD6bC8t2Y<...encrypted blob...>

            and reference it within CFNgin (NOTE: the path should be relative
            to the CFNgin config file)::

                conf_key: ${kms file://kms_value.txt}

                # Both of the above would resolve to
                conf_key: PASSWORD

        """
        value = read_value_from_path(value)

        region = None
        if "@" in value:
            region, value = value.split("@", 1)

        kms = get_session(region).client('kms')

        # encode str value as an utf-8 bytestring for use with codecs.decode.
        value = value.encode('utf-8')

        # get raw but still encrypted value from base64 version.
        decoded = codecs.decode(value, 'base64')

        # decrypt and return the plain text raw value.
        return kms.decrypt(CiphertextBlob=decoded)["Plaintext"]



"""Tests for runway.cfngin.environment."""
import unittest

from runway.cfngin.environment import parse_environment

TEST_ENV = """key1: value1
# some: comment

 # here: about

# key2
key2: value2

# another comment here
key3: some:complex::value


# one more here as well
key4: :otherValue:
key5: <another>@value
"""

TEST_ERROR_ENV = """key1: value1
error
"""


class TestEnvironment(unittest.TestCase):
    """Tests for runway.cfngin.environment."""

    def test_simple_key_value_parsing(self):
        """Test simple key value parsing."""
        parsed_env = parse_environment(TEST_ENV)
        self.assertTrue(isinstance(parsed_env, dict))
        self.assertEqual(parsed_env["key1"], "value1")
        self.assertEqual(parsed_env["key2"], "value2")
        self.assertEqual(parsed_env["key3"], "some:complex::value")
        self.assertEqual(parsed_env["key4"], ":otherValue:")
        self.assertEqual(parsed_env["key5"], "<another>@value")
        self.assertEqual(len(parsed_env), 5)

    def test_simple_key_value_parsing_exception(self):
        """Test simple key value parsing exception."""
        with self.assertRaises(ValueError):
            parse_environment(TEST_ERROR_ENV)

    def test_blank_value(self):
        """Test blank value."""
        env = """key1:"""
        parsed = parse_environment(env)
        self.assertEqual(parsed["key1"], "")



"""CFNgin session caching."""
import logging

import boto3

from .ui import ui

LOGGER = logging.getLogger(__name__)


# A global credential cache that can be shared among boto3 sessions. This is
# inherently threadsafe thanks to the GIL:
# https://docs.python.org/3/glossary.html#term-global-interpreter-lock
CREDENTIAL_CACHE = {}

DEFAULT_PROFILE = None


def get_session(region, profile=None):
    """Create a boto3 session or get a matching session from the cache.

    Args:
        region (str): The region for the session.
        profile (str): The profile for the session.

    Returns:
        :class:`boto3.session.Session`: A boto3 session with
        credential caching.

    """
    if profile is None:
        LOGGER.debug("No AWS profile explicitly provided. "
                     "Falling back to default.")
        profile = DEFAULT_PROFILE

    LOGGER.debug("Building session using profile \"%s\" in region \"%s\"",
                 profile, region)

    session = boto3.Session(region_name=region, profile_name=profile)
    cred_provider = session._session.get_component('credential_provider')
    provider = cred_provider.get_provider('assume-role')
    provider.cache = CREDENTIAL_CACHE
    provider._prompter = ui.getpass
    return session



"""Testing that the repository specified is downloaded to the correct directory."""

import os

from integration_tests.test_sources.test_sources import TestSources
from runway.util import change_dir
from util import run_command


class TestGitRemoteRepositoryCloned(TestSources):
    """Test deploying two regions in series."""

    TEST_NAME = __name__

    def __init__(self, logger):
        """Init class."""
        TestSources.__init__(self, logger)
        self.logger = logger

    def deploy(self):
        """Deploy provider."""
        self.copy_runway('git')
        with change_dir(self.sources_test_dir):
            return run_command(['runway', 'deploy'])

    def run(self):
        """Run tests."""
        self.clean()
        self.set_environment('dev')
        self.deploy()
        with change_dir(self.sources_test_dir):
            self.logger.info(os.path.isdir('.runway_cache'))
            assert os.path.isdir('.runway_cache')

        change_dir(self.base_dir)

    def teardown(self):
        """Teardown scaffolding."""
        self.set_env_var('CI', '1')
        self.logger.info('Tearing down: %s', self.TEST_NAME)
        with change_dir(self.sources_test_dir):
            run_command(['runway', 'destroy'])
        self.clean()
        self.unset_env_var('CI')



"""Retrieve a variable from the variables file or definition.

If the lookup is unable to find an defined variable matching the
provided query, the default value is returned or a ``ValueError`` is raised
if a default value was not provided.

Nested values can be used by providing the full path to the value but, it
will not select a list element.

The returned value can contain any YAML support data type
(dictionaries/mappings/hashes, lists/arrays/sequences, strings, numbers,
and booleon).

.. Example
.. code-block:: yaml

  deployment:
    - modules:
        - path: sampleapp.cfn
          environment:
            ami_id: ${var ami_id.${env AWS_REGION}}
      env_vars:
        SOME_VARIABLE: ${var some_variable::default=default}

"""
# pylint: disable=arguments-differ
from typing import Any, TYPE_CHECKING  # pylint: disable=unused-import

import logging

from .base import LookupHandler

# python2 supported pylint sees this is cyclic even though its only for type checking
# pylint: disable=cyclic-import
if TYPE_CHECKING:
    from ...config import VariablesDefinition  # noqa: F401 pylint: disable=unused-import
    from ...context import Context  # noqa: F401 pylint: disable=unused-import


LOGGER = logging.getLogger('runway')
TYPE_NAME = 'var'


class VarLookup(LookupHandler):
    """Variable definition lookup."""

    @classmethod
    def handle(cls, value, context, **kwargs):
        # type: (str, 'Context', Any) -> Any
        """Retrieve a variable from the variable definition.

        The value is retrieved from the variables passed to Runway using
        either a variables file or the ``variables`` directive of the
        config file.

        Args:
            value: The value passed to the lookup.
            variables: The resolved variables pass to Runway.

        Raises:
            ValueError: Unable to find a value for the provided query and
                a default value was not provided.

        """
        query, args = cls.parse(value)
        variables = kwargs['variables']

        result = variables.find(query, default=args.pop('default', ''))

        if result != '':  # allows for False bool and NoneType results
            return cls.transform(result, to_type=args.pop('transform', ''),
                                 **args)

        raise ValueError('"{}" does not exist in the variable definition')



"""CFNgin test."""
# pylint: disable=invalid-name
from os.path import basename

from integration_tests.test_cfngin.test_cfngin import Cfngin

FILE_BASENAME = '.'.join(basename(__file__).split('.')[:-1])


class TestNoStacksInConfig(Cfngin):
    """Test CFNgin with no stacks in config file."""

    REQUIRED_FIXTURE_FILES = [
        FILE_BASENAME + '.yaml'
    ]
    TEST_NAME = __name__

    def run(self):
        """Run the test."""
        self.copy_fixtures()
        code, _stdout, stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'
        expected_lines = [
            'WARNING: No stacks detected (error in config?)'
        ]
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def teardown(self):
        """Teardown any created resources and delete files."""
        self.cleanup_fixtures()



"""Tests for runway.cfngin.hooks.utils."""
# pylint: disable=unused-argument
import queue
import unittest

from runway.cfngin.config import Hook
from runway.cfngin.hooks.utils import handle_hooks

from ..factories import mock_context, mock_provider

HOOK_QUEUE = queue.Queue()


class TestHooks(unittest.TestCase):
    """Tests for runway.cfngin.hooks.utils."""

    def setUp(self):
        """Run before tests."""
        self.context = mock_context(namespace="namespace")
        self.provider = mock_provider(region="us-east-1")

    def test_empty_hook_stage(self):
        """Test empty hook stage."""
        hooks = []
        handle_hooks("fake", hooks, self.provider, self.context)
        self.assertTrue(HOOK_QUEUE.empty())

    def test_missing_required_hook(self):
        """Test missing required hook."""
        hooks = [Hook({"path": "not.a.real.path", "required": True})]
        with self.assertRaises(ImportError):
            handle_hooks("missing", hooks, self.provider, self.context)

    def test_missing_required_hook_method(self):
        """Test missing required hook method."""
        hooks = [{"path": "runway.cfngin.hooks.blah", "required": True}]
        with self.assertRaises(AttributeError):
            handle_hooks("missing", hooks, self.provider, self.context)

    def test_missing_non_required_hook_method(self):
        """Test missing non required hook method."""
        hooks = [Hook({"path": "runway.cfngin.hooks.blah", "required": False})]
        handle_hooks("missing", hooks, self.provider, self.context)
        self.assertTrue(HOOK_QUEUE.empty())

    def test_default_required_hook(self):
        """Test default required hook."""
        hooks = [Hook({"path": "runway.cfngin.hooks.blah"})]
        with self.assertRaises(AttributeError):
            handle_hooks("missing", hooks, self.provider, self.context)

    def test_valid_hook(self):
        """Test valid hook."""
        hooks = [
            Hook({"path": "tests.cfngin.hooks.test_utils.mock_hook",
                  "required": True})]
        handle_hooks("missing", hooks, self.provider, self.context)
        good = HOOK_QUEUE.get_nowait()
        self.assertEqual(good["provider"].region, "us-east-1")
        with self.assertRaises(queue.Empty):
            HOOK_QUEUE.get_nowait()

    def test_valid_enabled_hook(self):
        """Test valid enabled hook."""
        hooks = [
            Hook({"path": "tests.cfngin.hooks.test_utils.mock_hook",
                  "required": True, "enabled": True})]
        handle_hooks("missing", hooks, self.provider, self.context)
        good = HOOK_QUEUE.get_nowait()
        self.assertEqual(good["provider"].region, "us-east-1")
        with self.assertRaises(queue.Empty):
            HOOK_QUEUE.get_nowait()

    def test_valid_enabled_false_hook(self):
        """Test valid enabled false hook."""
        hooks = [
            Hook({"path": "tests.cfngin.hooks.test_utils.mock_hook",
                  "required": True, "enabled": False})]
        handle_hooks("missing", hooks, self.provider, self.context)
        self.assertTrue(HOOK_QUEUE.empty())

    def test_context_provided_to_hook(self):
        """Test context provided to hook."""
        hooks = [
            Hook({"path": "tests.cfngin.hooks.test_utils.context_hook",
                  "required": True})]
        handle_hooks("missing", hooks, "us-east-1", self.context)

    def test_hook_failure(self):
        """Test hook failure."""
        hooks = [
            Hook({"path": "tests.cfngin.hooks.test_utils.fail_hook",
                  "required": True})]
        with self.assertRaises(SystemExit):
            handle_hooks("fail", hooks, self.provider, self.context)
        hooks = [{"path": "tests.cfngin.hooks.test_utils.exception_hook",
                  "required": True}]
        with self.assertRaises(Exception):
            handle_hooks("fail", hooks, self.provider, self.context)
        hooks = [
            Hook({"path": "tests.cfngin.hooks.test_utils.exception_hook",
                  "required": False})]
        # Should pass
        handle_hooks("ignore_exception", hooks, self.provider, self.context)

    def test_return_data_hook(self):
        """Test return data hook."""
        hooks = [
            Hook({
                "path": "tests.cfngin.hooks.test_utils.result_hook",
                "data_key": "my_hook_results"
            }),
            # Shouldn't return data
            Hook({
                "path": "tests.cfngin.hooks.test_utils.context_hook"
            })
        ]
        handle_hooks("result", hooks, "us-east-1", self.context)

        self.assertEqual(
            self.context.hook_data["my_hook_results"]["foo"],
            "bar"
        )
        # Verify only the first hook resulted in stored data
        self.assertEqual(
            list(self.context.hook_data.keys()), ["my_hook_results"]
        )

    def test_return_data_hook_duplicate_key(self):
        """Test return data hook duplicate key."""
        hooks = [
            Hook({
                "path": "tests.cfngin.hooks.test_utils.result_hook",
                "data_key": "my_hook_results"
            }),
            Hook({
                "path": "tests.cfngin.hooks.test_utils.result_hook",
                "data_key": "my_hook_results"
            })
        ]

        with self.assertRaises(KeyError):
            handle_hooks("result", hooks, "us-east-1", self.context)

    def test_resolve_lookups_in_args(self):
        """Test the resolution of lookups in hook args."""
        hooks = [Hook({
            "path": "tests.cfngin.hooks.test_utils.kwargs_hook",
            "data_key": "my_hook_results",
            "args": {
                "default_lookup": "${default env_var::default_value}"
            }
        })]
        handle_hooks("lookups", hooks, "us-east-1", self.context)

        self.assertEqual(
            self.context.hook_data["my_hook_results"]["default_lookup"],
            "default_value"
        )


def mock_hook(*args, **kwargs):
    """Mock hook."""
    HOOK_QUEUE.put(kwargs)
    return True


def fail_hook(*args, **kwargs):
    """Fail hook."""
    return None


def exception_hook(*args, **kwargs):
    """Exception hook."""
    raise Exception


def context_hook(*args, **kwargs):
    """Context hook."""
    return "context" in kwargs


def result_hook(*args, **kwargs):
    """Results hook."""
    return {"foo": "bar"}


def kwargs_hook(*args, **kwargs):
    """Kwargs hook."""
    return kwargs



"""Tests for runway.cfngin.lookups.handlers.dynamodb."""
import unittest

import boto3
import mock
from botocore.stub import Stubber

from runway.cfngin.lookups.handlers.dynamodb import DynamodbLookup

from ...factories import SessionStub


class TestDynamoDBHandler(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.dynamodb.DynamodbLookup."""

    client = boto3.client('dynamodb', region_name='us-east-1',
                          # bypass the need to have these in the env
                          aws_access_key_id='testing',
                          aws_secret_access_key='testing')

    def setUp(self):
        """Run before tests."""
        self.stubber = Stubber(self.client)
        self.get_parameters_response = {'Item': {'TestMap': {'M': {
            'String1': {'S': 'StringVal1'},
            'List1': {'L': [
                {'S': 'ListVal1'},
                {'S': 'ListVal2'}]},
            'Number1': {'N': '12345'}, }}}}

    @mock.patch('runway.cfngin.lookups.handlers.dynamodb.get_session',
                return_value=SessionStub(client))
    def test_dynamodb_handler(self, _mock_client):
        """Test DynamoDB handler."""
        expected_params = {
            'TableName': 'TestTable',
            'Key': {
                'TestKey': {'S': 'TestVal'}
            },
            'ProjectionExpression': 'TestVal,TestMap,String1'
        }
        base_lookup_key = 'TestTable@TestKey:TestVal.TestMap[M].String1'
        base_lookup_key_valid = 'StringVal1'
        self.stubber.add_response('get_item',
                                  self.get_parameters_response,
                                  expected_params)
        with self.stubber:
            value = DynamodbLookup.handle(base_lookup_key)
            self.assertEqual(value, base_lookup_key_valid)

    @mock.patch('runway.cfngin.lookups.handlers.dynamodb.get_session',
                return_value=SessionStub(client))
    def test_dynamodb_number_handler(self, _mock_client):
        """Test DynamoDB number handler."""
        expected_params = {
            'TableName': 'TestTable',
            'Key': {
                'TestKey': {'S': 'TestVal'}
            },
            'ProjectionExpression': 'TestVal,TestMap,Number1'
        }
        base_lookup_key = 'TestTable@TestKey:TestVal.' \
            'TestMap[M].Number1[N]'
        base_lookup_key_valid = 12345
        self.stubber.add_response('get_item',
                                  self.get_parameters_response,
                                  expected_params)
        with self.stubber:
            value = DynamodbLookup.handle(base_lookup_key)
            self.assertEqual(value, base_lookup_key_valid)

    @mock.patch('runway.cfngin.lookups.handlers.dynamodb.get_session',
                return_value=SessionStub(client))
    def test_dynamodb_list_handler(self, _mock_client):
        """Test DynamoDB list handler."""
        expected_params = {
            'TableName': 'TestTable',
            'Key': {
                'TestKey': {'S': 'TestVal'}
            },
            'ProjectionExpression': 'TestVal,TestMap,List1'
        }
        base_lookup_key = 'TestTable@TestKey:TestVal.' \
            'TestMap[M].List1[L]'
        base_lookup_key_valid = ['ListVal1', 'ListVal2']
        self.stubber.add_response('get_item',
                                  self.get_parameters_response,
                                  expected_params)
        with self.stubber:
            value = DynamodbLookup.handle(base_lookup_key)
            self.assertEqual(value, base_lookup_key_valid)

    @mock.patch('runway.cfngin.lookups.handlers.dynamodb.get_session',
                return_value=SessionStub(client))
    def test_dynamodb_empty_table_handler(self, _mock_client):
        """Test DynamoDB empty table handler."""
        expected_params = {
            'TableName': '',
            'Key': {
                'TestKey': {'S': 'TestVal'}
            },
            'ProjectionExpression': 'TestVal,TestMap,String1'
        }
        base_lookup_key = '@TestKey:TestVal.TestMap[M].String1'
        self.stubber.add_response('get_item',
                                  self.get_parameters_response,
                                  expected_params)
        with self.stubber:
            try:
                DynamodbLookup.handle(base_lookup_key)
            except ValueError as err:
                self.assertEqual(
                    'Please make sure to include a DynamoDB table name',
                    str(err))

    @mock.patch('runway.cfngin.lookups.handlers.dynamodb.get_session',
                return_value=SessionStub(client))
    def test_dynamodb_missing_table_handler(self, _mock_client):
        """Test DynamoDB missing table handler."""
        expected_params = {
            'Key': {
                'TestKey': {'S': 'TestVal'}
            },
            'ProjectionExpression': 'TestVal,TestMap,String1'
        }
        base_lookup_key = 'TestKey:TestVal.TestMap[M].String1'
        self.stubber.add_response('get_item',
                                  self.get_parameters_response,
                                  expected_params)
        with self.stubber:
            try:
                DynamodbLookup.handle(base_lookup_key)
            except ValueError as err:
                self.assertEqual(
                    'Please make sure to include a tablename',
                    str(err))

    @mock.patch('runway.cfngin.lookups.handlers.dynamodb.get_session',
                return_value=SessionStub(client))
    def test_dynamodb_invalid_table_handler(self, _mock_client):
        """Test DynamoDB invalid table handler."""
        expected_params = {
            'TableName': 'FakeTable',
            'Key': {
                'TestKey': {'S': 'TestVal'}
            },
            'ProjectionExpression': 'TestVal,TestMap,String1'
        }
        base_lookup_key = 'FakeTable@TestKey:TestVal.TestMap[M].String1'
        service_error_code = 'ResourceNotFoundException'
        self.stubber.add_client_error('get_item',
                                      service_error_code=service_error_code,
                                      expected_params=expected_params)
        with self.stubber:
            try:
                DynamodbLookup.handle(base_lookup_key)
            except ValueError as err:
                self.assertEqual(
                    'Cannot find the DynamoDB table: FakeTable',
                    str(err))

    @mock.patch('runway.cfngin.lookups.handlers.dynamodb.get_session',
                return_value=SessionStub(client))
    def test_dynamodb_invalid_partition_key_handler(self, _mock_client):
        """Test DynamoDB invalid partition key handler."""
        expected_params = {
            'TableName': 'TestTable',
            'Key': {
                'FakeKey': {'S': 'TestVal'}
            },
            'ProjectionExpression': 'TestVal,TestMap,String1'
        }
        base_lookup_key = 'TestTable@FakeKey:TestVal.TestMap[M].String1'
        service_error_code = 'ValidationException'
        self.stubber.add_client_error('get_item',
                                      service_error_code=service_error_code,
                                      expected_params=expected_params)

        with self.stubber:
            try:
                DynamodbLookup.handle(base_lookup_key)
            except ValueError as err:
                self.assertEqual(
                    'No DynamoDB record matched the partition key: FakeKey',
                    str(err))

    @mock.patch('runway.cfngin.lookups.handlers.dynamodb.get_session',
                return_value=SessionStub(client))
    def test_dynamodb_invalid_partition_val_handler(self, _mock_client):
        """Test DynamoDB invalid partition val handler."""
        expected_params = {
            'TableName': 'TestTable',
            'Key': {
                'TestKey': {'S': 'FakeVal'}
            },
            'ProjectionExpression': 'FakeVal,TestMap,String1'
        }
        empty_response = {'ResponseMetadata': {}}
        base_lookup_key = 'TestTable@TestKey:FakeVal.TestMap[M].String1'
        self.stubber.add_response('get_item',
                                  empty_response,
                                  expected_params)
        with self.stubber:
            try:
                DynamodbLookup.handle(base_lookup_key)
            except ValueError as err:
                self.assertEqual(
                    'The DynamoDB record could not be found using '
                    'the following key: {\'S\': \'FakeVal\'}',
                    str(err))



"""Tests for runway.cfngin.stacker."""
import unittest

from runway.cfngin.commands import Stacker
from runway.cfngin.exceptions import InvalidConfig


class TestStacker(unittest.TestCase):
    """Tests for runway.cfngin.stacker.Stacker."""

    def test_stacker_build_parse_args(self):
        """Test stacker build parse args."""
        stacker = Stacker()
        args = stacker.parse_args(
            ["build",
             "-r", "us-west-2",
             "-e", "namespace=test.override",
             "tests/cfngin/fixtures/basic.env",
             "tests/cfngin/fixtures/vpc-bastion-db-web.yaml"]
        )
        self.assertEqual(args.region, "us-west-2")
        self.assertFalse(args.outline)
        # verify namespace was modified
        self.assertEqual(args.environment["namespace"], "test.override")

    def test_stacker_build_parse_args_region_from_env(self):
        """Test stacker build parse args region from env."""
        stacker = Stacker()
        args = stacker.parse_args(
            ["build",
             "-e", "namespace=test.override",
             "tests/cfngin/fixtures/basic.env",
             "tests/cfngin/fixtures/vpc-bastion-db-web.yaml"]
        )
        self.assertEqual(args.region, None)

    def test_stacker_build_context_passed_to_blueprint(self):
        """Test stacker build context passed to blueprint."""
        stacker = Stacker()
        args = stacker.parse_args(
            ["build",
             "-r", "us-west-2",
             "tests/cfngin/fixtures/basic.env",
             "tests/cfngin/fixtures/vpc-bastion-db-web.yaml"]
        )
        stacker.configure(args)
        stacks_dict = args.context.get_stacks_dict()
        blueprint = stacks_dict[args.context.get_fqn("bastion")].blueprint
        self.assertTrue(hasattr(blueprint, "context"))
        blueprint.render_template()
        # verify that the bastion blueprint only contains blueprint variables,
        # not BaseDomain, AZCount or CidrBlock. Any variables that get passed
        # in from the command line shouldn't be resolved at the blueprint level
        self.assertNotIn("BaseDomain", blueprint.template.parameters)
        self.assertNotIn("AZCount", blueprint.template.parameters)
        self.assertNotIn("CidrBlock", blueprint.template.parameters)

    def test_stacker_blueprint_property_access_does_not_reset_blueprint(self):
        """Test stacker blueprint property access does not reset blueprint."""
        stacker = Stacker()
        args = stacker.parse_args(
            ["build",
             "-r", "us-west-2",
             "tests/cfngin/fixtures/basic.env",
             "tests/cfngin/fixtures/vpc-bastion-db-web.yaml"]
        )
        stacker.configure(args)
        stacks_dict = args.context.get_stacks_dict()
        bastion_stack = stacks_dict[args.context.get_fqn("bastion")]
        bastion_stack.blueprint.render_template()
        self.assertIn("DefaultSG", bastion_stack.blueprint.template.parameters)

    def test_stacker_build_context_stack_names_specified(self):
        """Test stacker build context stack names specified."""
        stacker = Stacker()
        args = stacker.parse_args(
            ["build",
             "-r", "us-west-2",
             "tests/cfngin/fixtures/basic.env",
             "tests/cfngin/fixtures/vpc-bastion-db-web.yaml",
             "--stacks", "vpc",
             "--stacks", "bastion"]
        )
        stacker.configure(args)
        stacks = args.context.get_stacks()
        self.assertEqual(len(stacks), 2)

    def test_stacker_build_fail_when_parameters_in_stack_def(self):
        """Test stacker build fail when parameters in stack def."""
        stacker = Stacker()
        args = stacker.parse_args(
            ["build",
             "-r", "us-west-2",
             "tests/cfngin/fixtures/basic.env",
             "tests/cfngin/fixtures/vpc-bastion-db-web-pre-1.0.yaml"]
        )
        with self.assertRaises(InvalidConfig):
            stacker.configure(args)

    def test_stacker_build_custom_info_log_format(self):
        """Test stacker build custom info log format."""
        stacker = Stacker()
        args = stacker.parse_args(
            [
                "build", "-r", "us-west-2",
                "tests/cfngin/fixtures/not-basic.env",
                "tests/cfngin/fixtures/vpc-custom-log-format-info.yaml"
            ]
        )
        stacker.configure(args)
        self.assertEqual(
            stacker.config.log_formats["info"],
            '[%(asctime)s] test custom log format - %(message)s'
        )
        # for some reason, pylint does not see DictType.get as valid
        self.assertIsNone(
            stacker.config.log_formats.get("color")  # pylint: disable=no-member
        )
        self.assertIsNone(
            stacker.config.log_formats.get("debug")  # pylint: disable=no-member
        )


if __name__ == '__main__':
    unittest.main()



"""Tests for the Source type object."""
import logging
import unittest

from runway.sources.git import Git

LOGGER = logging.getLogger('runway')


class GitTester(unittest.TestCase):
    """Tests for the Source type object."""

    def test_fetch_returns_directory_string(self):
        """Ensure a directory string is returned."""
        fetched = Git(**{
            'options': {},
            'uri': 'git://github.com/onicagroup/runway.git',
            'location': '/'
        }).fetch()
        self.assertEqual(fetched, '/')

    def test_sanitize_git_path(self):
        """Ensure git path is property sanitized"""
        path = Git().sanitize_git_path('git://github.com/onicagroup/runway.git')
        self.assertEqual(path, 'github.com_onicagroup_runway')



"""Tests for lookup handler for env."""
from unittest import TestCase

from runway.lookups.handlers.env import EnvLookup
from runway.util import MutableMap


CONTEXT = MutableMap(**{
    'env_vars': {
        'str_val': 'test'
    }
})


class TestEnvLookup(TestCase):
    """Tests for EnvLookup."""

    def test_handle(self):
        """Validate handle base functionality."""
        query = 'str_val'
        result = EnvLookup.handle(query, context=CONTEXT)

        self.assertEqual(result, 'test')

    def test_handle_not_found(self):
        """Validate exception when lookup cannot be resolved."""
        query = 'NOT_VALID'

        with self.assertRaises(ValueError):
            EnvLookup.handle(query, context=CONTEXT)



"""Packaging settings."""
from codecs import open as codecs_open
from os.path import abspath, dirname, join

from setuptools import find_packages, setup

from runway import __version__


THIS_DIR = abspath(dirname(__file__))
with codecs_open(join(THIS_DIR, 'README.md'), encoding='utf-8') as readfile:
    LONG_DESCRIPTION = readfile.read()


INSTALL_REQUIRES = [
    'Send2Trash',
    'awacs',  # for embedded hooks
    # awscli included for embedded hooks and aws subcommand
    'awscli>=1.16.308<2.0',
    'backports.tempfile; python_version < "3.2"',
    'botocore>=1.12.111',  # matching awscli/boto3 requirement
    'boto3>=1.9.111<2.0',
    'cfn_flip>=1.2.1',  # 1.2.1+ require PyYAML 4.1+
    'cfn-lint',
    'docker',
    'docopt',
    'requests',
    'future',
    'pyhcl~=0.4',
    'gitpython',
    'pyOpenSSL',  # For embedded hook & associated script usage
    'PyYAML>=4.1,<5.3',  # match awscli top-end
    'six>=1.13.0',
    'typing;python_version<"3.5"',
    'yamllint',
    'zgitignore',  # for embedded hooks
    'troposphere>=2.4.2',
    # botocore pins its urllib3 dependency like this, so we need to do the
    # same to ensure v1.25+ isn't pulled in by pip
    'urllib3>=1.20,<1.25',
    # dependency of importlib-metadata, dependency of pytest, cfn-lint, & others
    # 2.0.0 drops support for python 3.5
    'zipp~=1.0.0',
    # inherited from stacker 1.7.0 requirements
    'jinja2>=2.7,<3.0',
    'schematics>=2.0.1,<2.1.0',
    'formic2'
]


setup(
    name='runway',
    version=__version__,
    description='Simplify infrastructure/app testing/deployment',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url='https://github.com/onicagroup/runway',
    author='Onica Group LLC',
    author_email='opensource@onica.com',
    license='Apache License 2.0',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=2.6',
    keywords='cli',
    packages=find_packages(exclude=('tests', 'integration_tests')),
    install_requires=INSTALL_REQUIRES,
    entry_points={
        'console_scripts': [
            'runway=runway.cli:main',
        ],
    },
    scripts=['scripts/stacker-runway', 'scripts/stacker-runway.cmd',
             'scripts/tf-runway', 'scripts/tf-runway.cmd',
             'scripts/tfenv-runway', 'scripts/tfenv-runway.cmd'],
    include_package_data=True  # needed for templates,blueprints,hooks
)



"""CFNgin context."""
import collections
import json
import logging

from .config import Config
from .exceptions import (PersistentGraphCannotLock,
                         PersistentGraphCannotUnlock,
                         PersistentGraphLockCodeMissmatch,
                         PersistentGraphLocked, PersistentGraphUnlocked)
from .plan import Graph
from .session_cache import get_session
from .stack import Stack
from .target import Target
from .util import ensure_s3_bucket

LOGGER = logging.getLogger(__name__)


DEFAULT_NAMESPACE_DELIMITER = "-"
DEFAULT_TEMPLATE_INDENT = 4


def get_fqn(base_fqn, delimiter, name=None):
    """Return the fully qualified name of an object within this context.

    If the name passed already appears to be a fully qualified name, it
    will be returned with no further processing.

    """
    if name and name.startswith("%s%s" % (base_fqn, delimiter)):
        return name

    return delimiter.join([_f for _f in [base_fqn, name] if _f])


class Context(object):
    """The context under which the current stacks are being executed.

    The CFNgin Context is responsible for translating the values passed in
    via the command line and specified in the config to `Stack` objects.

    """

    def __init__(self, environment=None,
                 stack_names=None,
                 config=None,
                 region=None,
                 force_stacks=None):
        """Instantiate class.

        Args:
            environment (dict): A dictionary used to pass in information about
                the environment. Useful for templating.
            stack_names (list): A list of stack_names to operate on. If not
                passed, usually all stacks defined in the config will be
                operated on.
            config (:class:`runway.cfngin.config.Config`): The CFNgin
                configuration being operated on.
            region (str): Name of an AWS region if provided as a CLI argument.
            force_stacks (list): A list of stacks to force work on. Used to
                work on locked stacks.

        """
        self._bucket_name = None
        self._persistent_graph = None
        self._persistent_graph_lock_code = None
        self._persistent_graph_lock_tag = 'cfngin_lock_code'
        self._s3_bucket_verified = None
        self._stacks = None
        self._targets = None
        self._upload_to_s3 = None
        self.config = config or Config()
        self.bucket_region = self.config.cfngin_bucket_region or region
        self.environment = environment
        self.force_stacks = force_stacks or []
        self.hook_data = {}
        self.s3_conn = get_session(self.bucket_region).client('s3')
        self.stack_names = stack_names or []

    @property
    def _base_fqn(self):
        """Return ``namespace`` sanitized for use as an S3 Bucket name."""
        return self.namespace.replace(".", "-").lower()

    @property
    def _persistent_graph_tags(self):
        """Cache of tags on the persistent graph object.

        Returns:
            Dict[str, str]

        """
        try:
            return {t['Key']: t['Value'] for t in
                    self.s3_conn.get_object_tagging(
                        **self.persistent_graph_location).get('TagSet', {})}
        except self.s3_conn.exceptions.NoSuchKey:
            LOGGER.debug('Persistant graph object does not exist in S3; '
                         'could not get tags')
            return {}

    @property
    def bucket_name(self):
        """Return ``cfngin_bucket`` from config, calculated name, or None."""
        if not self.upload_to_s3:
            return None

        return self.config.cfngin_bucket \
            or "stacker-%s" % (self.get_fqn(),)

    @property
    def mappings(self):
        """Return ``mappings`` from config."""
        return self.config.mappings or {}

    @property
    def namespace(self):
        """Return ``namespace`` from config."""
        return self.config.namespace

    @property
    def namespace_delimiter(self):
        """Return ``namespace_delimiter`` from config or default."""
        delimiter = self.config.namespace_delimiter
        if delimiter is not None:
            return delimiter
        return DEFAULT_NAMESPACE_DELIMITER

    @property
    def persistent_graph(self):
        """Graph if a persistent graph is being used.

        Will create an "empty" object in S3 if one is not found.

        Returns:
            :class:`runway.cfngin.plan.Graph`

        """
        if not self._persistent_graph:
            if not self.persistent_graph_location:
                return None

            content = '{}'

            if self.s3_bucket_verified:
                try:
                    LOGGER.debug('Getting persistent graph from s3:\n%s',
                                 json.dumps(self.persistent_graph_location,
                                            indent=4))
                    content = self.s3_conn.get_object(
                        ResponseContentType='application/json',
                        **self.persistent_graph_location
                    )['Body'].read().decode('utf-8')
                except self.s3_conn.exceptions.NoSuchKey:
                    LOGGER.info('Persistant graph object does not exist '
                                'in S3; creating one now.')
                    self.s3_conn.put_object(
                        Body=content,
                        ServerSideEncryption='AES256',
                        ACL='bucket-owner-full-control',
                        ContentType='application/json',
                        **self.persistent_graph_location
                    )
            self.persistent_graph = json.loads(content)

        return self._persistent_graph

    @persistent_graph.setter
    def persistent_graph(self, graph_dict):
        """Load a persistent graph dict as a :class:`runway.cfngin.plan.Graph`."""
        self._persistent_graph = Graph.from_dict(graph_dict, self)

    @property
    def persistent_graph_location(self):
        """Location of the persistent graph in s3.

        Returns:
            Dict[str, str] Bucket and Key for the object in S3.

        """
        if not self.upload_to_s3 or not self.config.persistent_graph_key:
            return {}

        return {
            'Bucket': self.bucket_name,
            'Key': 'persistent_graphs/{namespace}/{key}'.format(
                namespace=self.config.namespace,
                key=(self.config.persistent_graph_key + '.json' if not
                     self.config.persistent_graph_key.endswith('.json')
                     else self.config.persistent_graph_key)
            )
        }

    @property
    def persistent_graph_lock_code(self):
        """Code used to lock the persistent graph S3 object.

        Returns:
            Optional[str]

        """
        if (not self._persistent_graph_lock_code and
                self.persistent_graph_location):
            self._persistent_graph_lock_code = self._persistent_graph_tags.get(
                self._persistent_graph_lock_tag
            )
        return self._persistent_graph_lock_code

    @property
    def persistent_graph_locked(self):
        """Check if persistent graph is locked.

        Returns:
            bool

        """
        if not self.persistent_graph:
            return False
        if not self.persistent_graph_lock_code:
            return False
        return True

    @property
    def s3_bucket_verified(self):
        """Check CFNgin bucket exists and you have access.

        If the CFNgin bucket does not exist, will try to create one.

        Returns:
            bool

        """
        if not self._s3_bucket_verified and self.bucket_name:
            ensure_s3_bucket(self.s3_conn,
                             self.bucket_name,
                             self.bucket_region,
                             persist_graph=bool(self.persistent_graph_location))
            self._s3_bucket_verified = True
        return self._s3_bucket_verified

    @property
    def tags(self):
        """Return ``tags`` from config."""
        tags = self.config.tags
        if tags is not None:
            return tags
        if self.namespace:
            return {"cfngin_namespace": self.namespace}
        return {}

    @property
    def template_indent(self):
        """Return ``template_indent`` from config or default."""
        indent = self.config.template_indent
        if indent is not None:
            return int(indent)
        return DEFAULT_TEMPLATE_INDENT

    @property
    def upload_to_s3(self):
        """Check if S3 should be used for caching/persistent graph.

        Returns:
            (bool)

        """
        if not self._upload_to_s3:
            # Don't upload stack templates to S3 if `cfngin_bucket` is
            # explicitly set to an empty string.
            if self.config.cfngin_bucket == '':
                LOGGER.debug("Not uploading to s3 because `cfngin_bucket` "
                             "is explicitly set to an empty string")
                return False

            # If no namespace is specificied, and there's no explicit
            # cfngin bucket specified, don't upload to s3. This makes
            # sense because we can't realistically auto generate a cfngin
            # bucket name in this case.
            if not self.namespace and not self.config.cfngin_bucket:
                LOGGER.debug("Not uploading to s3 because there is no "
                             "namespace set, and no cfngin_bucket set")
                return False

        return True

    def _get_stack_definitions(self):
        """Return ``stacks`` from config."""
        return self.config.stacks

    def get_targets(self):
        """Return the named targets that are specified in the config.

        Returns:
            list: a list of :class:`runway.cfngin.target.Target` objects

        """
        if not self._targets:
            targets = []
            for target_def in self.config.targets or []:
                target = Target(target_def)
                targets.append(target)
            self._targets = targets
        return self._targets

    def get_stack(self, name):
        """Get a stack by name.

        Args:
            name (str): Name of a stack to retrieve.

        """
        for stack in self.get_stacks():
            if stack.name == name:
                return stack
        return None

    def get_stacks(self):
        """Get the stacks for the current action.

        Handles configuring the :class:`runway.cfngin.stack.Stack` objects
        that will be used in the current action.

        Returns:
            list: a list of :class:`runway.cfngin.stack.Stack` objects

        """
        if not self._stacks:
            stacks = []
            definitions = self._get_stack_definitions()
            for stack_def in definitions:
                stack = Stack(
                    definition=stack_def,
                    context=self,
                    mappings=self.mappings,
                    force=stack_def.name in self.force_stacks,
                    locked=stack_def.locked,
                    enabled=stack_def.enabled,
                    protected=stack_def.protected,
                )
                stacks.append(stack)
            self._stacks = stacks
        return self._stacks

    def get_stacks_dict(self):
        """Construct a dict of {stack.fqn: stack} for easy access to stacks."""
        return dict((stack.fqn, stack) for stack in self.get_stacks())

    def get_fqn(self, name=None):
        """Return the fully qualified name of an object within this context.

        If the name passed already appears to be a fully qualified name, it
        will be returned with no further processing.

        """
        return get_fqn(self._base_fqn, self.namespace_delimiter, name)

    def lock_persistent_graph(self, lock_code):
        """Locks the persistent graph in s3.

        Args:
            lock_code (str): The code that will be used to lock the S3 object.

        Raises:
            :class:`runway.cfngin.exceptions.PersistentGraphLocked`
            :class:`runway.cfngin.exceptions.PersistentGraphCannotLock`

        """
        if not self.persistent_graph:
            return

        if self.persistent_graph_locked:
            raise PersistentGraphLocked

        try:
            self.s3_conn.put_object_tagging(
                Tagging={'TagSet': [
                    {'Key': self._persistent_graph_lock_tag,
                     'Value': lock_code}
                ]},
                **self.persistent_graph_location
            )
            LOGGER.info('Locked persistent graph "%s" with lock ID "%s".',
                        '/'.join([self.persistent_graph_location['Bucket'],
                                  self.persistent_graph_location['Key']]),
                        lock_code)
        except self.s3_conn.exceptions.NoSuchKey:
            raise PersistentGraphCannotLock('s3 object does not exist')

    def put_persistent_graph(self, lock_code):
        """Upload persistent graph to s3.

        Args:
            lock_code (str): The code that will be used to lock the S3 object.

        Raises:
            :class:`runway.cfngin.exceptions.PersistentGraphUnlocked`
            :class:`runway.cfngin.exceptions.PersistentGraphLockCodeMissmatch`

        """
        if not self.persistent_graph:
            return

        if not self.persistent_graph.to_dict():
            self.s3_conn.delete_object(**self.persistent_graph_location)
            LOGGER.debug('Removed empty persistent graph object from S3')
            return

        if not self.persistent_graph_locked:
            raise PersistentGraphUnlocked(
                reason='It must be locked by the current session to be '
                       'updated.'
            )

        if self.persistent_graph_lock_code != lock_code:
            raise PersistentGraphLockCodeMissmatch(
                lock_code, self.persistent_graph_lock_code
            )

        self.s3_conn.put_object(
            Body=self.persistent_graph.dumps(4),
            ServerSideEncryption='AES256',
            ACL='bucket-owner-full-control',
            ContentType='application/json',
            Tagging='{}={}'.format(self._persistent_graph_lock_tag,
                                   lock_code),
            **self.persistent_graph_location
        )
        LOGGER.debug('Persistent graph updated:\n%s',
                     self.persistent_graph.dumps(indent=4))

    def set_hook_data(self, key, data):
        """Set hook data for the given key.

        Args:
            key(str): The key to store the hook data in.
            data(:class:`collections.Mapping`): A dictionary of data to store,
                as returned from a hook.

        """
        if not isinstance(data, collections.Mapping):
            raise ValueError("Hook (key: %s) data must be an instance of "
                             "collections.Mapping (a dictionary for "
                             "example)." % key)

        if key in self.hook_data:
            raise KeyError("Hook data for key %s already exists, each hook "
                           "must have a unique data_key." % key)

        self.hook_data[key] = data

    def unlock_persistent_graph(self, lock_code):
        """Unlocks the persistent graph in s3.

        Args:
            lock_code (str): The code that will be used to lock the S3 object.

        Raises:
            :class:`runway.cfngin.exceptions.PersistentGraphCannotUnlock`

        """
        if not self.persistent_graph:
            return True

        if not self.persistent_graph.to_dict():
            try:
                self.s3_conn.get_object(
                    ResponseContentType='application/json',
                    **self.persistent_graph_location
                )
            except self.s3_conn.exceptions.NoSuchKey:
                LOGGER.info('Persistent graph was deleted and does not '
                            'need to be unlocked.')
                return True

        LOGGER.debug('Unlocking persistent graph "%s".',
                     self.persistent_graph_location)

        if not self.persistent_graph_locked:
            raise PersistentGraphCannotUnlock(PersistentGraphUnlocked(
                reason='It must be locked by the current session to be '
                       'unlocked.'
            ))

        if self.persistent_graph_lock_code == lock_code:
            try:
                self.s3_conn.delete_object_tagging(
                    **self.persistent_graph_location
                )
            except self.s3_conn.exceptions.NoSuchKey:
                pass
            self._persistent_graph_lock_code = None
            LOGGER.info('Unlocked persistent graph "%s".',
                        '/'.join([self.persistent_graph_location['Bucket'],
                                  self.persistent_graph_location['Key']]))
            return True
        raise PersistentGraphCannotUnlock(
            PersistentGraphLockCodeMissmatch(
                lock_code,
                self.persistent_graph_lock_code
            )
        )



"""Tests for runway.cfngin.lookups.handlers.hook_data."""
import unittest

from runway.cfngin.context import Context
from runway.cfngin.lookups.handlers.hook_data import HookDataLookup


class TestHookDataLookup(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.hook_data.HookDataLookup."""

    def setUp(self):
        """Run before tests."""
        self.ctx = Context({"namespace": "test-ns"})
        self.ctx.set_hook_data("fake_hook", {"result": "good"})

    def test_valid_hook_data(self):
        """Test valid hook data."""
        value = HookDataLookup.handle("fake_hook::result", context=self.ctx)
        self.assertEqual(value, "good")

    def test_invalid_hook_data(self):
        """Test invalid hook data."""
        with self.assertRaises(KeyError):
            HookDataLookup.handle("fake_hook::bad_key", context=self.ctx)

    def test_bad_value_hook_data(self):
        """Test bad value hook data."""
        with self.assertRaises(ValueError):
            HookDataLookup.handle("fake_hook", context=self.ctx)



"""Hook utils."""
import collections
import logging
import os
import sys

from runway.util import load_object_from_string
from runway.variables import Variable, resolve_variables

from ..exceptions import FailedVariableLookup

LOGGER = logging.getLogger(__name__)


def full_path(path):
    """Return full path."""
    return os.path.abspath(os.path.expanduser(path))


def handle_hooks(stage, hooks, provider, context):
    """Handle pre/post_build hooks.

    These are pieces of code that we want to run before/after the builder
    builds the stacks.

    Args:
        stage (str): The current stage (pre_run, post_run, etc).
        hooks (List[:class:`runway.cfngin.config.Hook`]): Hooks to execute.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`): Provider
            instance.
        context (:class:`runway.cfngin.context.Context`): Context instance.

    """
    if not hooks:
        LOGGER.debug("No %s hooks defined.", stage)
        return

    hook_paths = []
    for i, hook in enumerate(hooks):
        try:
            hook_paths.append(hook.path)
        except KeyError:
            raise ValueError("%s hook #%d missing path." % (stage, i))

    LOGGER.info("Executing %s hooks: %s", stage, ", ".join(hook_paths))
    for hook in hooks:
        data_key = hook.data_key
        required = hook.required

        if not hook.enabled:
            LOGGER.debug("hook with method %s is disabled, skipping",
                         hook.path)
            continue

        try:
            method = load_object_from_string(hook.path)
        except (AttributeError, ImportError):
            LOGGER.exception("Unable to load method at %s:", hook.path)
            if required:
                raise
            continue

        if isinstance(hook.args, dict):
            args = [Variable(k, v) for k, v in hook.args.items()]
            try:  # handling for output or similar being used in pre_build
                resolve_variables(args, context, provider)
            except FailedVariableLookup:
                if 'pre' in stage:
                    LOGGER.error('Lookups that change the order of '
                                 'execution, like "output", can only be '
                                 'used in "post_*" hooks. Please '
                                 'ensure that the hook being used does '
                                 'not rely on a stack, hook_data, or '
                                 'context that does not exist yet.')
                raise
            kwargs = {v.name: v.value for v in args}
        else:
            kwargs = hook.args or {}

        try:
            result = method(context=context, provider=provider, **kwargs)
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception("Method %s threw an exception:", hook.path)
            if required:
                raise
            continue

        if not result:
            if required:
                LOGGER.error("Required hook %s failed. Return value: %s",
                             hook.path, result)
                sys.exit(1)
            LOGGER.warning("Non-required hook %s failed. Return value: %s",
                           hook.path, result)
        else:
            if isinstance(result, collections.Mapping):
                if data_key:
                    LOGGER.debug("Adding result for hook %s to context in "
                                 "data_key %s.", hook.path, data_key)
                    context.set_hook_data(data_key, result)
                else:
                    LOGGER.debug("Hook %s returned result data, but no data "
                                 "key set, so ignoring.", hook.path)



"""CFNgin exceptions."""


class CancelExecution(Exception):
    """Raised when we want to cancel executing the plan."""


class ChangesetDidNotStabilize(Exception):
    """Raised when the applying a changeset fails."""

    def __init__(self, change_set_id):
        """Instantiate class.

        Args:
            change_set_id (str): The changeset that failed.

        """
        self.id = change_set_id
        message = "Changeset '%s' did not reach a completed state." % (
            change_set_id
        )

        super(ChangesetDidNotStabilize, self).__init__(message)


class FailedLookup(Exception):
    """Intermediary Exception to be converted to FailedVariableLookup.

    Should be caught by error handling and
    :class:`runway.cfngin.exceptions.FailedVariableLookup` raised instead to
    construct a propper error message.

    """

    def __init__(self, lookup, error, *args, **kwargs):
        """Instantiate class.

        Args:
            lookup (:class:`runway.cfngin.variables.VariableValueLookup`):
                Attempted lookup and resulted in an exception being raised.
            error (Exception): The exception that was raised.

        """
        self.lookup = lookup
        self.error = error
        super(FailedLookup, self).__init__("Failed lookup", *args, **kwargs)


class FailedVariableLookup(Exception):
    """Lookup could not be resolved.

    Raised when an exception is raised when trying to resolve a lookup.

    """

    def __init__(self, variable_name, lookup, error, *args, **kwargs):
        """Instantiate class.

        Args:
            variable_name (str): Name of the variable that failed to be
                resolved.
            lookup (:class:`runway.cfngin.variables.VariableValueLookup`):
                Attempted lookup and resulted in an exception being raised.
            error (Exception): The exception that was raised.

        """
        self.lookup = lookup
        self.error = error
        message = "Couldn't resolve lookup in variable `%s`, " % variable_name
        message += "lookup: ${%s}: " % repr(lookup)
        message += "(%s) %s" % (error.__class__, error)
        super(FailedVariableLookup, self).__init__(message, *args, **kwargs)


class GraphError(Exception):
    """Raised when the graph is invalid (e.g. acyclic dependencies)."""

    def __init__(self, exception, stack, dependency):
        """Instantiate class.

        Args:
            exception (Exception): The exception that was raised by the invalid
                graph.
            stack (str): Name of the stack causing the error.
            dependency (str): Name of the dependency causing the error.

        """
        self.stack = stack
        self.dependency = dependency
        self.exception = exception
        message = (
            "Error detected when adding '%s' "
            "as a dependency of '%s': %s"
        ) % (dependency, stack, str(exception))
        super(GraphError, self).__init__(message)


class ImproperlyConfigured(Exception):
    """Raised when a componenet is improperly configured."""

    def __init__(self, cls, error, *args, **kwargs):
        """Instantiate class.

        Args:
            cls (Any): The class that was improperly configured.
            error (Exception): The exception that was raised when trying to
                use cls.
        """
        message = "Class \"%s\" is improperly configured: %s" % (
            cls,
            error,
        )
        super(ImproperlyConfigured, self).__init__(message, *args, **kwargs)


class InvalidConfig(Exception):
    """Provided config file is invalid."""

    def __init__(self, errors):
        """Instantiate class.

        Args:
            errors (Union[str, List[Union[Exception, str]]]): Errors or error
                messages that are raised to identify that a config is invalid.

        """
        super(InvalidConfig, self).__init__(errors)
        self.errors = errors


class InvalidDockerizePipConfiguration(Exception):
    """Raised when the provided configuration for dockerized pip is invalid."""

    def __init__(self, msg):
        """Instantiate class.

        Args:
            msg (str): The reason for the error being raised.

        """
        self.message = msg
        super(InvalidDockerizePipConfiguration, self).__init__(self.message)


class InvalidLookupCombination(Exception):
    """Improper use of lookups to result in a non-string return value."""

    def __init__(self, lookup, lookups, value, *args, **kwargs):
        """Instantiate class.

        Args:
            lookup (:class:`runway.cfngin.variables.VariableValueLookup`): The
                variable lookup that was attempted but did not return a string.
            lookups (:class:`runway.cfngin.variables.VariableValueConcatenation`):
                The full variable concatenation the failing lookup is part of.
            value (Any): The non-string value returned by lookup.

        """
        message = (
            "Lookup: \"{}\" has non-string return value, must be only lookup "
            "present (not {}) in \"{}\""
        ).format(str(lookup), len(lookups), value)
        super(InvalidLookupCombination, self).__init__(message,
                                                       *args,
                                                       **kwargs)


class InvalidLookupConcatenation(Exception):
    """Intermediary Exception to be converted to InvalidLookupCombination.

    Should be caught by error handling and
    :class:`runway.cfngin.exceptions.InvalidLookupCombination` raised instead
    to construct a propper error message.

    """

    def __init__(self, lookup, lookups, *args, **kwargs):
        """Instantiate class."""
        self.lookup = lookup
        self.lookups = lookups
        super(InvalidLookupConcatenation, self).__init__("", *args, **kwargs)


class InvalidUserdataPlaceholder(Exception):
    """Raised when a placeholder name in raw_user_data is not valid.

    E.g ``${100}`` would raise this.

    """

    def __init__(self, blueprint_name, exception_message, *args, **kwargs):
        """Instantiate class.

        Args:
            blueprint_name (str): Name of the blueprint with invalid userdata
                placeholder.
            exception_message (str): Message from the exception that was raised
                while parsing the userdata.

        """
        message = exception_message + ". "
        message += "Could not parse userdata in blueprint \"%s\". " % (
            blueprint_name)
        message += "Make sure to escape all $ symbols with a $$."
        super(InvalidUserdataPlaceholder, self).__init__(
            message, *args, **kwargs)


class MissingEnvironment(Exception):
    """Raised when an environment lookup is used but the key doesn't exist."""

    def __init__(self, key, *args, **kwargs):
        """Instantiate class.

        Args:
            key (str): The key that was used but doesn't exist in the
            environment.

        """
        self.key = key
        message = "Environment missing key %s." % (key,)
        super(MissingEnvironment, self).__init__(message, *args, **kwargs)


class MissingParameterException(Exception):
    """Raised if a required parameter with no default is missing."""

    def __init__(self, parameters, *args, **kwargs):
        """Instantiate class.

        Args:
            parameters (List[str]): A list of the parameters that are missing.

        """
        self.parameters = parameters
        message = "Missing required cloudformation parameters: %s" % (
            ", ".join(parameters),
        )
        super(MissingParameterException, self).__init__(message, *args,
                                                        **kwargs)


class MissingVariable(Exception):
    """Raised when a variable with no default is not provided a value."""

    def __init__(self, blueprint_name, variable_name, *args, **kwargs):
        """Instantiate class.

        Args:
            blueprint_name (str): Name of the blueprint.
            variable_name (str): Name of the variable missing a value.

        """
        message = "Variable \"%s\" in blueprint \"%s\" is missing" % (
            variable_name, blueprint_name)
        super(MissingVariable, self).__init__(message, *args, **kwargs)


class OutputDoesNotExist(Exception):
    """Raised when a specific stack output does not exist."""

    def __init__(self, stack_name, output, *args, **kwargs):
        """Instantiate class.

        Args:
            stack_name (str): Name of the stack.
            output (str): The output that does not exist.

        """
        self.stack_name = stack_name
        self.output = output

        message = "Output %s does not exist on stack %s" % (output,
                                                            stack_name)
        super(OutputDoesNotExist, self).__init__(message, *args, **kwargs)


class PipError(Exception):
    """Raised when pip returns a non-zero exit code."""

    def __init__(self):
        """Instantiate class."""
        self.message = ('A non-zero exit code was returned when invoking '
                        'pip. More information can be found in the error '
                        'above.')
        super(PipError, self).__init__(self.message)


class PipenvError(Exception):
    """Raised when pipenv returns a non-zero exit code."""

    def __init__(self):
        """Instantiate class."""
        self.message = ('A non-zero exit code was returned when invoking '
                        'pipenv. Please ensure pipenv in installed and the '
                        'Pipfile being used is valid. More information can be '
                        'found in the error above.')
        super(PipenvError, self).__init__(self.message)


class PersistentGraphCannotLock(Exception):
    """Raised when the persistent graph in S3 cannot be locked."""

    def __init__(self, reason):
        """Instantiate class."""
        message = "Could not lock persistent graph; %s" % reason
        super(PersistentGraphCannotLock, self).__init__(message)


class PersistentGraphCannotUnlock(Exception):
    """Raised when the persistent graph in S3 cannot be unlocked."""

    def __init__(self, reason):
        """Instantiate class."""
        message = "Could not unlock persistent graph; %s" % reason
        super(PersistentGraphCannotUnlock, self).__init__(message)


class PersistentGraphLocked(Exception):
    """Raised when the persistent graph in S3 is lock.

    The action being executed requires it to be unlocked before attempted.

    """

    def __init__(self, message=None, reason=None):
        """Instantiate class."""
        if not message:
            message = ("Persistant graph is locked. {}".format(
                reason or ("This action requires the graph to be "
                           "unlocked to be executed.")
            ))
        super(PersistentGraphLocked, self).__init__(message)


class PersistentGraphLockCodeMissmatch(Exception):
    """Raised when the provided persistent graph lock code does not match.

    The code used to unlock the persistent graph must match the s3 object lock
    code.

    """

    def __init__(self, provided_code, s3_code):
        """Instantiate class."""
        message = ("The provided lock code '%s' does not match the S3 "
                   "object lock code '%s'" % (provided_code, s3_code))
        super(PersistentGraphLockCodeMissmatch, self).__init__(message)


class PersistentGraphUnlocked(Exception):
    """Raised when the persistent graph in S3 is unlock.

    The action being executed requires it to be locked before attempted.

    """

    def __init__(self, message=None, reason=None):
        """Instantiate class."""
        if not message:
            message = ("Persistant graph is unlocked. {}".format(
                reason or ("This action requires the graph to be "
                           "locked to be executed.")
            ))
        super(PersistentGraphUnlocked, self).__init__(message)


class PlanFailed(Exception):
    """Raised if any step of a plan fails."""

    def __init__(self, failed_steps, *args, **kwargs):
        """Instantiate class.

        Args:
            failed_steps (List[:class:`runway.cfngin.plan.Step`]): The steps
                that failed.

        """
        self.failed_steps = failed_steps

        step_names = ', '.join(step.name for step in failed_steps)
        message = "The following steps failed: %s" % (step_names,)

        super(PlanFailed, self).__init__(message, *args, **kwargs)


class StackDidNotChange(Exception):
    """Raised when there are no changes to be made by the provider."""


class StackDoesNotExist(Exception):
    """Raised when a stack does not exist in AWS."""

    def __init__(self, stack_name, *args, **kwargs):
        """Instantiate class.

        Args:
            stack_name (str): Name of the stack that does not exist.

        """
        message = ("Stack: \"%s\" does not exist in outputs or the lookup is "
                   "not available in this CFNgin run") % (stack_name,)
        super(StackDoesNotExist, self).__init__(message, *args, **kwargs)


class StackUpdateBadStatus(Exception):
    """Raised if the state of a stack can't be handled."""

    def __init__(self, stack_name, stack_status, reason, *args, **kwargs):
        """Instantiate class.

        Args:
            stack_name (str): Name of the stack.
            stack_status (str): The stack's status.
            reason (str): The reason for the current status.

        """
        self.stack_name = stack_name
        self.stack_status = stack_status

        message = ("Stack: \"%s\" cannot be updated nor re-created from state "
                   "%s: %s" % (stack_name, stack_status, reason))
        super(StackUpdateBadStatus, self).__init__(message, *args, **kwargs)


class UnableToExecuteChangeSet(Exception):
    """Raised if changeset execution status is not ``AVAILABLE``."""

    def __init__(self, stack_name, change_set_id, execution_status):
        """Instantiate class.

        Args:
            stack_name (str): Name of the stack.
            change_set_id (str): The changeset that failed.
            execution_status (str): The value of the changeset's
                ``ExecutionStatus`` attribute.

        """
        self.stack_name = stack_name
        self.id = change_set_id
        self.execution_status = execution_status

        message = ("Changeset '%s' on stack '%s' had bad execution status: "
                   "%s" % (change_set_id, stack_name, execution_status))

        super(UnableToExecuteChangeSet, self).__init__(message)


class UnhandledChangeSetStatus(Exception):
    """Raised when creating a changeset failed for an unhandled reason.

    Handled failure reasons include: no changes

    """

    def __init__(self, stack_name, change_set_id, status, status_reason):
        """Instantiate class.

        Args:
            stack_name (str): Name of the stack.
            change_set_id (str): The changeset that failed.
            status (str): The state that could not be handled.
            status_reason (str): Cause of the current state.

        """
        self.stack_name = stack_name
        self.id = change_set_id
        self.status = status
        self.status_reason = status_reason
        message = (
            "Changeset '%s' on stack '%s' returned an unhandled status "
            "'%s: %s'." % (change_set_id, stack_name, status,
                           status_reason)
        )

        super(UnhandledChangeSetStatus, self).__init__(message)


class UnknownLookupType(Exception):
    """Lookup type provided does not match a registered lookup.

    Example:
        If a lookup of ``${<lookup_type> query}`` is used and ``<lookup_type>``
        is not a registered lookup, this exception will be raised.

    """

    def __init__(self, lookup_type, *args, **kwargs):
        """Instantiate class.

        Args:
            lookup_type (str): Lookup type that was used but not registered.

        """
        message = "Unknown lookup type: \"{}\"".format(lookup_type)
        super(UnknownLookupType, self).__init__(message, *args, **kwargs)


class UnresolvedVariable(Exception):
    """Raised when trying to use a variable before it has been resolved."""

    def __init__(self, blueprint_name, variable, *args, **kwargs):
        """Instantiate class.

        Args:
            blueprint_name (str): Name of the blueprint that tried to use
                the unresolved variables.
            variable (:class:`runway.cfngin.variables.Variable`): The
                unresolved variable.

        """
        message = (
            "Variable \"%s\" in blueprint \"%s\" hasn't been resolved" % (
                variable.name, blueprint_name
            )
        )
        super(UnresolvedVariable, self).__init__(message, *args, **kwargs)


class UnresolvedVariableValue(Exception):
    """Intermediary Exception to be converted to UnresolvedVariable.

    Should be caught by error handling and
    :class:`runway.cfngin.exceptions.UnresolvedVariable` raised instead to
    construct a propper error message.

    """

    def __init__(self, lookup, *args, **kwargs):
        """Instantiate class.

        Args:
            lookup (:class:`runway.cfngin.variables.VariableValueLookup`): The
                lookup that is not resolved.

        """
        self.lookup = lookup
        super(UnresolvedVariableValue, self).__init__(
            "Unresolved lookup", *args, **kwargs)


class UnresolvedVariables(Exception):
    """Raised when trying to use variables before they has been resolved."""

    def __init__(self, blueprint_name, *args, **kwargs):
        """Instantiate class.

        Args:
            blueprint_name (str): Name of the blueprint that tried to use
                the unresolved variables.

        """
        message = "Blueprint: \"%s\" hasn't resolved it's variables" % (
            blueprint_name)
        super(UnresolvedVariables, self).__init__(message, *args, **kwargs)


class ValidatorError(Exception):
    """Used for errors raised by custom validators of blueprint variables."""

    def __init__(self, variable, validator, value, exception=None):
        """Instantiate class.

        Args:
            variable (str): The variable that failed validation.
            validator (str): The validator that was not passed.
            value (str): The value of the variable that did not pass the
                validator.
            exception (Exception): The exception raised by the validator.

        """
        self.variable = variable
        self.validator = validator
        self.value = value
        self.exception = exception
        self.message = ("Validator '%s' failed for variable '%s' with value "
                        "'%s'") % (self.validator, self.variable, self.value)

        if self.exception:
            self.message += ": %s: %s" % (self.exception.__class__.__name__,
                                          str(self.exception))
        super(ValidatorError, self).__init__()

    def __str__(self):
        """Return the exception's message when converting to a string."""
        return self.message


class VariableTypeRequired(Exception):
    """Raised when a variable defined in a blueprint is missing a type."""

    def __init__(self, blueprint_name, variable_name, *args, **kwargs):
        """Instantiate class.

        Args:
            blueprint_name (str): Name of the blueprint.
            variable_name (str): Name of the variable missing a type.

        """
        message = (
            "Variable \"%s\" in blueprint \"%s\" does not have a type" % (
                variable_name, blueprint_name)
        )
        super(VariableTypeRequired, self).__init__(message, *args, **kwargs)
